import numpy

import chainer
from chainer import backend
from chainer.backends import cuda
from chainer import configuration
from chainer import function_node
from chainer.utils import type_check


if cuda.cudnn_enabled:
    cudnn = cuda.cudnn
    libcudnn = cuda.cuda.cudnn


class GroupNormalization(function_node.FunctionNode):

    def __init__(self, groups, eps=1e-5):
        if not isinstance(groups, int):
            raise TypeError('Argument: \'groups\' type must be (int).')

        self.groups = groups
        self.eps = eps
        self.mean = None
        self.inv_std = None
        self.dummy_gamma = None

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 3)
        x_type, gamma_type, beta_type = in_types
        type_check.expect(
            x_type.dtype.kind == 'f',
            x_type.ndim >= 2,
            gamma_type.ndim == 1,
            beta_type.ndim == 1,
            gamma_type.dtype.kind == 'f',
            gamma_type.dtype == beta_type.dtype,
            x_type.shape[1] == gamma_type.shape[0],
            gamma_type.shape == beta_type.shape,
        )

    def forward(self, inputs):
        if inputs[0].shape[1] % self.groups != 0:
            raise ValueError('The number of channels {} is not divisible by '
                             '\'groups\' argument {}.'
                             .format(inputs[0].shape[1], self.groups))

        xp = backend.get_array_module(*inputs)
        if xp is cuda.cupy and chainer.should_use_cudnn('>=auto', 5000):
            return self.forward_cudnn(inputs)

        self.retain_inputs((0, 1))
        x, gamma, beta = inputs

        interm_dtype = numpy.promote_types(x.dtype, gamma.dtype)

        gamma = gamma.astype(interm_dtype, copy=False)
        beta = beta.astype(interm_dtype, copy=False)

        orig_shape = x.shape
        batch_size, channels = orig_shape[:2]
        groups = self.groups
        reduced_shape = (batch_size * groups, -1)
        x = x.reshape(reduced_shape)

        self.mean = x.mean(axis=1, dtype=interm_dtype)
        x_hat = x - self.mean[:, None]
        var = (x_hat * x_hat).mean(axis=1)
        var += self.eps
        self.inv_std = var
        del var
        xp.sqrt(self.inv_std, out=self.inv_std)
        xp.reciprocal(self.inv_std, out=self.inv_std)
        x_hat *= self.inv_std[:, None]

        y = x_hat.reshape((batch_size, channels, -1))
        y *= gamma[:, None]
        y += beta[:, None]

        y = y.reshape(orig_shape)
        return y.astype(x.dtype, copy=False),

    def forward_cudnn(self, inputs):
        if self.eps < libcudnn.CUDNN_BN_MIN_EPSILON:
            raise RuntimeError(
                'cuDNN does not allow an eps value '
                'less than {}.'.format(libcudnn.CUDNN_BN_MIN_EPSILON))

        self.retain_inputs((0, 1))
        x, gamma, beta = inputs
        xp = cuda.cupy

        interm_dtype = numpy.promote_types(x.dtype, gamma.dtype)

        gamma = gamma.astype(interm_dtype, copy=False)
        beta = beta.astype(interm_dtype, copy=False)

        orig_shape = x.shape
        batch_size, channels = orig_shape[:2]
        groups = self.groups
        cudnn_shape = (1, batch_size * groups, -1, 1)
        x = x.reshape(cudnn_shape)

        with x.device:
            dummy_beta = xp.zeros(batch_size * groups, dtype=beta.dtype)
            self.dummy_gamma = xp.ones_like(dummy_beta)
        x_hat, self.mean, self.inv_std = \
            cudnn.batch_normalization_forward_training(
                x, self.dummy_gamma, dummy_beta, dummy_beta, dummy_beta, None,
                None, self.eps, 1.0, True, libcudnn.CUDNN_BATCHNORM_SPATIAL,
                configuration.config.debug)

        y = x_hat.reshape((batch_size, channels, -1))
        cuda.elementwise(
            'T gamma, T beta', 'U y',
            'y = y * gamma + beta',
            'groupnorm_y')(gamma[:, None], beta[:, None], y)

        y = y.reshape(orig_shape)
        return y,

    def backward(self, indexes, grad_outputs):
        x, gamma = self.get_retained_inputs()
        gy, = grad_outputs

        interm_dtype = numpy.promote_types(x.dtype, gamma.dtype)

        gamma = chainer.functions.cast(gamma, interm_dtype)

        orig_shape = x.shape
        batch_size = orig_shape[0]
        groups = self.groups
        reduced_shape = (batch_size * groups, -1)
        x = x.reshape(reduced_shape)
        x_ = chainer.functions.cast(x, interm_dtype)

        x_hat, = _XHat(
            self.eps, self.mean, self.inv_std,
            self.dummy_gamma).apply((x_,))
        gx_hat, ggamma, gbeta = _ScaleShiftGrad().apply((
            x_hat, gamma, chainer.functions.cast(gy, interm_dtype)))
        gx, = _XHatGrad(
            self.eps, self.mean, self.inv_std,
            self.dummy_gamma, x_hat.array).apply(
                (x_, gx_hat))

        gx = gx.reshape(orig_shape)
        return chainer.functions.cast(gx, x.dtype), ggamma, gbeta


class _ScaleShiftGrad(function_node.FunctionNode):

    def forward(self, inputs):
        self.retain_inputs((0, 1, 2))
        x_hat, gamma, gy = inputs

        batch_size, channels = gy.shape[:2]
        gy = gy.reshape((batch_size, channels, -1))

        reduced_shape = x_hat.shape
        x_hat = x_hat.reshape((batch_size, channels, -1))

        gx_hat = gy * gamma[:, None]
        gbeta = gy.sum(axis=(0, 2))
        if backend.get_array_module(x_hat) is cuda.cupy:
            ggamma = cuda.reduce(
                'T gy, T x_hat', 'T ggamma',
                'gy * x_hat', 'a + b', 'ggamma = a', '0',
                'groupnorm_ggamma')(gy, x_hat, axis=(0, 2))
        else:
            ggamma = (gy * x_hat).sum(axis=(0, 2))

        gx_hat = gx_hat.reshape(reduced_shape)
        return gx_hat, ggamma, gbeta

    def backward(self, indexes, grad_outputs):
        x_hat, gamma, gy = self.get_retained_inputs()
        ggx_hat, gggamma, ggbeta = grad_outputs

        orig_shape = gy.shape
        batch_size, channels = gy.shape[:2]
        gy = gy.reshape((batch_size, channels, -1))

        reduced_shape = x_hat.shape
        x_hat = x_hat.reshape((batch_size, channels, -1))
        ggx_hat = ggx_hat.reshape((batch_size, channels, -1))

        gx_hat2 = gggamma[:, None] * gy
        ggamma2 = chainer.functions.sum(ggx_hat * gy, axis=(0, 2))
        ggy = (ggx_hat * gamma[:, None] + gggamma[:, None] * x_hat +
               ggbeta[:, None])

        gx_hat2 = gx_hat2.reshape(reduced_shape)
        ggy = ggy.reshape(orig_shape)
        return gx_hat2, ggamma2, ggy


class _XHat(function_node.FunctionNode):

    def __init__(self, eps, mean, inv_std, dummy_gamma):
        self.eps = eps
        self.mean = mean
        self.inv_std = inv_std
        self.dummy_gamma = dummy_gamma

    def forward_cpu(self, inputs):
        self.retain_inputs((0,))
        x, = inputs
        x_hat = x - self.mean[:, None]
        x_hat *= self.inv_std[:, None]
        self.retain_outputs((0,))
        return x_hat,

    def forward_gpu(self, inputs):
        self.retain_inputs((0,))
        x, = inputs
        x_hat = cuda.elementwise(
            'T x, U mean, U inv_std', 'T x_hat',
            'x_hat = (x - mean) * inv_std',
            'groupnorm_x_hat')(x, self.mean[:, None], self.inv_std[:, None])
        self.retain_outputs((0,))
        return x_hat,

    def backward(self, indexes, grad_outputs):
        x, = self.get_retained_inputs()
        x_hat, = self.get_retained_outputs()
        gx_hat, = grad_outputs
        return _XHatGrad(
            self.eps, self.mean, self.inv_std,
            self.dummy_gamma, x_hat.array).apply((x, gx_hat))


class _XHatGrad(function_node.FunctionNode):

    def __init__(self, eps, mean, inv_std, dummy_gamma, x_hat):
        self.eps = eps
        self.mean = mean
        self.inv_std = inv_std
        self.dummy_gamma = dummy_gamma
        self.x_hat = x_hat

    def forward(self, inputs):
        xp = backend.get_array_module(*inputs)
        if xp is cuda.cupy and chainer.should_use_cudnn('>=auto', 5000) and \
                self.dummy_gamma is not None:
            return self.forward_cudnn(inputs)

        self.retain_inputs((0, 1))
        _, gx_hat = inputs
        x_hat = self.x_hat
        self.x_hat = None

        gx_hat_avg = gx_hat.mean(axis=1, keepdims=True)
        gx_hat_x_hat_avg = (gx_hat * x_hat).mean(axis=1, keepdims=True)
        gx_std = gx_hat - gx_hat_avg - x_hat * gx_hat_x_hat_avg
        gx = self.inv_std[:, None] * gx_std

        self.retain_outputs((0,))
        return gx,

    def forward_cudnn(self, inputs):
        if self.eps < libcudnn.CUDNN_BN_MIN_EPSILON:
            raise RuntimeError(
                'cuDNN does not allow an eps value '
                'less than {}.'.format(libcudnn.CUDNN_BN_MIN_EPSILON))

        self.retain_inputs((0, 1))
        x, gx_hat = inputs
        self.x_hat = None

        # `x[None, :, :, None]` is slower because it results in a different
        # strides and cuDNN doesn't recognize it as a contiguous array.
        reduced_shape = x.shape
        cudnn_shape = (1,) + reduced_shape + (1,)
        x = x.reshape(cudnn_shape)
        gx_hat = gx_hat.reshape(cudnn_shape)

        gx, _, _ = cudnn.batch_normalization_backward(
            x, self.dummy_gamma, gx_hat,
            self.mean, self.inv_std, self.eps,
            True, libcudnn.CUDNN_BATCHNORM_SPATIAL,
            configuration.config.debug)
        gx = gx.reshape(reduced_shape)

        self.retain_outputs((0,))
        return gx,

    def backward(self, indexes, grad_outputs):
        F = chainer.functions
        x, gx_hat = self.get_retained_inputs()
        gx, = self.get_retained_outputs()
        ggx, = grad_outputs

        x_hat, = _XHat(
            self.eps, self.mean, self.inv_std,
            self.dummy_gamma).apply((x,))

        ret = []

        if 0 in indexes:
            # -- sketch of gx2, which is grad of x through gx
            # gx = inv_std * gx_std
            # dgx = dinv_std * gx_std + inv_std * dgx_std
            #
            # -gx2l = (ggx * dinv_std * gx_std) / dx
            #       = sum(ggx * gx_std) * (dinv_std / dx)
            #       = -sum(ggx * gx_std) * inv_std^2 * x_hat / N
            #       = -inv_std * x_hat * mean(ggx * gx)
            #
            # By `gx_std = gx_hat - gx_hat_avg - x_hat * gx_hat_x_hat_avg`,
            # -gx_hat2r = (ggx * inv_std * dgx_std) / dx_hat
            #           = -inv_std * (ggx * mean(gx_hat * x_hat) +
            #                         gx_hat * mean(ggx * x_hat))

            gx2l_std = x_hat * F.mean(ggx * gx, axis=1, keepdims=True)
            gx2l, = _MulInvStd(
                self.eps, self.mean, self.inv_std,
                self.dummy_gamma).apply((x, gx2l_std))

            gx_hat2r_std = (
                ggx * F.mean(gx_hat * x_hat, axis=1, keepdims=True) +
                gx_hat * F.mean(ggx * x_hat, axis=1, keepdims=True))
            gx_hat2r, = _MulInvStd(
                self.eps, self.mean, self.inv_std,
                self.dummy_gamma).apply((x, gx_hat2r_std))
            gx2r, = _XHatGrad(
                self.eps, self.mean, self.inv_std,
                self.dummy_gamma, x_hat.array).apply((x, gx_hat2r))

            gx2 = -(gx2l + gx2r)
            ret.append(gx2)

        if 1 in indexes:
            ggx_hat, = _XHatGrad(
                self.eps, self.mean, self.inv_std,
                self.dummy_gamma, x_hat.array).apply((x, ggx))
            ret.append(ggx_hat)

        return ret


class _MulInvStd(function_node.FunctionNode):

    def __init__(self, eps, mean, inv_std, dummy_gamma):
        self.eps = eps
        self.mean = mean
        self.inv_std = inv_std
        self.dummy_gamma = dummy_gamma

    def forward(self, inputs):
        self.retain_inputs((0,))
        _, y = inputs
        z = self.inv_std[:, None] * y
        self.retain_outputs((0,))
        return z,

    def backward(self, indexes, grad_outputs):
        x, = self.get_retained_inputs()
        z, = self.get_retained_outputs()
        gz, = grad_outputs

        x_hat, = _XHat(
            self.eps, self.mean, self.inv_std,
            self.dummy_gamma).apply((x,))
        gx_std = x_hat * chainer.functions.mean(gz * z, axis=1, keepdims=True)
        gx, = _MulInvStd(
            self.eps, self.mean, self.inv_std,
            self.dummy_gamma).apply((x, gx_std))
        gy, = _MulInvStd(
            self.eps, self.mean, self.inv_std,
            self.dummy_gamma).apply((x, gz))

        return -gx, gy


def group_normalization(x, groups, gamma, beta, eps=1e-5):
    """Group normalization function.

    This function implements a "group normalization"
    which divides the channels into groups and computes within each group
    the mean and variance, then normalize by these statistics,
    scales and shifts them.

    Args:
        x (:class:`~chainer.Variable` or :ref:`ndarray`): Batch tensors.
            First dimension of this value must be the size of minibatch and
            second dimension must be the number of channels.
            Moreover, this value must have one or more following dimensions,
            such as height and width.
        groups (int):
            The number of channel groups.
            This value must be a divisor of the number of channels.
        gamma (:class:`~chainer.Variable` or :ref:`ndarray`):
            Scaling parameter.
        beta (:class:`~chainer.Variable` or :ref:`ndarray`):
            Shifting parameter.
        eps (float): Epsilon value for numerical stability of normalization.

    Returns:
        ~chainer.Variable: The output variable which has the same shape
        as :math:`x`.

    See: `Group Normalization <https://arxiv.org/abs/1803.08494>`_

    .. seealso::

        :class:`~chainer.links.GroupNormalization` to manage the model
        parameters ``gamma`` and ``beta``.

    """
    return GroupNormalization(groups, eps).apply((x, gamma, beta))[0]



import six

import chainer
from chainer.dataset import dataset_mixin


class TabularDataset(dataset_mixin.DatasetMixin):
    """An abstract class that represents tabular dataset.

    This class represents a tabular dataset.
    In a tabular dataset, all examples have the same number of elements.
    For example, all examples of the dataset below have three elements
    (:obj:`a[i]`, :obj:`b[i]`, and :obj:`c[i]`).

    .. csv-table::
        :header: , a, b, c

        0, :obj:`a[0]`, :obj:`b[0]`, :obj:`c[0]`
        1, :obj:`a[1]`, :obj:`b[1]`, :obj:`c[1]`
        2, :obj:`a[2]`, :obj:`b[2]`, :obj:`c[2]`
        3, :obj:`a[3]`, :obj:`b[3]`, :obj:`c[3]`

    Since an example can be represented by both tuple and dict (
    :obj:`(a[i], b[i], c[i])` and :obj:`{'a': a[i], 'b': b[i], 'c': c[i]}`),
    this class uses :attr:`mode` to indicate which representation will be used.
    If there is only one column, an example also can be represented by a value
    (:obj:`a[i]`). In this case, :attr:`mode` is :obj:`None`.

    An inheritance should implement
    :meth:`__len__`, :attr:`keys`, :attr:`mode` and :meth:`get_examples`.

    >>> import numpy as np
    >>>
    >>> from chainer import dataset
    >>>
    >>> class MyDataset(dataset.TabularDataset):
    ...
    ...     def __len__(self):
    ...         return 4
    ...
    ...     @property
    ...     def keys(self):
    ...          return ('a', 'b', 'c')
    ...
    ...     @property
    ...     def mode(self):
    ...          return tuple
    ...
    ...     def get_examples(self, indices, key_indices):
    ...          data = np.arange(12).reshape((4, 3))
    ...          if indices is not None:
    ...              data = data[indices]
    ...          if key_indices is not None:
    ...              data = data[:, list(key_indices)]
    ...          return tuple(data.transpose())
    ...
    >>> dataset = MyDataset()
    >>> len(dataset)
    4
    >>> dataset.keys
    ('a', 'b', 'c')
    >>> dataset.astuple()[0]
    (0, 1, 2)
    >>> sorted(dataset.asdict()[0].items())
    [('a', 0), ('b', 1), ('c', 2)]
    >>>
    >>> view = dataset.slice[[3, 2], ('c', 0)]
    >>> len(view)
    2
    >>> view.keys
    ('c', 'a')
    >>> view.astuple()[1]
    (8, 6)
    >>> sorted(view.asdict()[1].items())
    [('a', 6), ('c', 8)]

    """

    def __len__(self):
        raise NotImplementedError

    @property
    def keys(self):
        """Names of columns.

        A tuple of strings that indicate the names of columns.
        """
        raise NotImplementedError

    @property
    def mode(self):
        """Mode of representation.

        This indicates the type of value returned
        by :meth:`fetch` and :meth:`__getitem__`.
        :class:`tuple`, :class:`dict`, and :obj:`None` are supported.
        """
        raise NotImplementedError

    def get_examples(self, indices, key_indices):
        """Return a part of data.

        Args:
            indices (list of ints or slice): Indices of requested rows.
                If this argument is :obj:`None`, it indicates all rows.
            key_indices (tuple of ints): Indices of requested columns.
                If this argument is :obj:`None`, it indicates all columns.

        Returns:
            tuple of lists/arrays
        """
        raise NotImplementedError

    @property
    def slice(self):
        """Get a slice of dataset.

        Args:
           indices (list/array of ints/bools or slice): Requested rows.
           keys (tuple of ints/strs or int or str): Requested columns.

        Returns:
            A view of specified range.
        """
        return chainer.dataset.tabular._slice._SliceHelper(self)

    def fetch(self):
        """Fetch data.

        This method fetches all data of the dataset/view.
        Note that this method returns a column-major data
        (i.e. :obj:`([a[0], ..., a[3]], ..., [c[0], ... c[3]])`,
        :obj:`{'a': [a[0], ..., a[3]], ..., 'c': [c[0], ..., c[3]]}`, or
        :obj:`[a[0], ..., a[3]]`).

        Returns:
            If :attr:`mode` is :class:`tuple`,
            this method returns a tuple of lists/arrays.
            If :attr:`mode` is :class:`dict`,
            this method returns a dict of lists/arrays.
        """
        examples = self.get_examples(None, None)
        if self.mode is tuple:
            return examples
        elif self.mode is dict:
            return dict(six.moves.zip(self.keys, examples))
        elif self.mode is None:
            return examples[0]

    def convert(self, data):
        """Convert fetched data.

        This method takes data fetched by :meth:`fetch` and
        pre-process them before passing them to models.
        The default behaviour is converting each column into an ndarray.
        This behaviour can be overridden by :meth:`with_converter`.
        If the dataset is constructed by :meth:`concat` or :meth:`join`,
        the converter of the first dataset is used.

        Args:
            data (tuple or dict): Data from :meth:`fetch`.

        Returns:
            A tuple or dict.
            Each value is an ndarray.
        """
        if isinstance(data, tuple):
            return tuple(_as_array(d) for d in data)
        elif isinstance(data, dict):
            return {k: _as_array(v) for k, v in data.items()}
        else:
            return _as_array(data)

    def astuple(self):
        """Return a view with tuple mode.

        Returns:
            A view whose :attr:`mode` is :class:`tuple`.
        """
        return chainer.dataset.tabular._asmode._Astuple(self)

    def asdict(self):
        """Return a view with dict mode.

        Returns:
            A view whose :attr:`mode` is :class:`dict`.
        """
        return chainer.dataset.tabular._asmode._Asdict(self)

    def concat(self, *datasets):
        """Stack datasets along rows.

        Args:
            datasets (iterable of :class:`TabularDataset`):
                Datasets to be concatenated.
                All datasets must have the same :attr:`keys`.

        Returns:
            A concatenated dataset.
        """
        return chainer.dataset.tabular._concat._Concat(self, *datasets)

    def join(self, *datasets):
        """Stack datasets along columns.

        Args:
            datasets (iterable of :class:`TabularDataset`):
                Datasets to be concatenated.
                All datasets must have the same length

        Returns:
            A joined dataset.
        """
        return chainer.dataset.tabular._join._Join(self, *datasets)

    def transform(self, keys, transform):
        """Apply a transform to each example.

        Args:
            keys (tuple of strs): The keys of transformed examples.
            transform (callable): A callable that takes an example
                and returns transformed example. :attr:`mode` of
                transformed dataset is determined by the transformed
                examples.

        Returns:
            A transfromed dataset.
        """
        return chainer.dataset.tabular._transform._Transform(
            self, keys, transform)

    def transform_batch(self, keys, transform_batch):
        """Apply a transform to examples.

        Args:
            keys (tuple of strs): The keys of transformed examples.
            transform_batch (callable): A callable that takes examples
                and returns transformed examples. :attr:`mode` of
                transformed dataset is determined by the transformed
                examples.

        Returns:
            A transfromed dataset.
        """
        return chainer.dataset.tabular._transform._TransformBatch(
            self, keys, transform_batch)

    def with_converter(self, converter):
        """Override the behaviour of :meth:`convert`.

        This method overrides :meth:`convert`.

        Args:
            converter (callable): A new converter.

        Returns:
            A dataset with the new converter.
        """

        return chainer.dataset.tabular._with_converter._WithConverter(
            self, converter)

    def get_example(self, i):
        example = self.get_examples([i], None)
        example = tuple(col[0] for col in example)
        if self.mode is tuple:
            return example
        elif self.mode is dict:
            return dict(six.moves.zip(self.keys, example))
        elif self.mode is None:
            return example[0]

    def __iter__(self):
        return (self.get_example(i) for i in six.moves.range(len(self)))


def _as_array(data):
    if isinstance(data, chainer.get_array_types()):
        return data
    else:
        device = chainer.backend.get_device_from_array(data[0])
        with chainer.using_device(device):
            return device.xp.asarray(data)



from __future__ import division

import numpy

from chainer import backend
from chainer.backends import cuda
from chainer import function_node
from chainer.utils import type_check


def _infer_lines(B, C, H, W, out_H, out_W, kH, kW):
    target_size = 2 ** 17
    line_size = B * C * (H * W // out_H + kH * kW * out_W)
    target_lines = target_size // line_size

    if target_lines < out_H:
        lines = 1
        while True:
            next_lines = lines * 2
            if next_lines > target_lines:
                break
            lines = next_lines
    else:
        lines = out_H

    return lines


def interpolate_bilinear_cpu(x, v, u, vw, uw):
    B, C, H, W = x.shape
    out_H, out_W = v.shape

    # Interpolation is done by each output panel (i.e. multi lines)
    # in order to better utilize CPU cache memory.
    lines = _infer_lines(B, C, H, W, out_H, out_W, 2, 2)

    vcol = numpy.empty((2, lines, out_W), dtype=v.dtype)
    ucol = numpy.empty((2, lines, out_W), dtype=u.dtype)
    wcol = numpy.empty((2, 2, lines, out_W), dtype=x.dtype)

    y = numpy.empty((B * C, out_H * out_W), dtype=x.dtype)

    for i in range(0, out_H, lines):
        l = min(lines, out_H - i)
        vcol = vcol[:, :l]
        ucol = ucol[:, :l]
        wcol = wcol[:, :, :l]
        i_end = i + l

        # indices
        vcol[0] = v[i:i_end]
        ucol[0] = u[i:i_end]
        numpy.add(vcol[0], 1, out=vcol[1])
        numpy.add(ucol[0], 1, out=ucol[1])
        numpy.minimum(vcol[1], H - 1, out=vcol[1])
        numpy.minimum(ucol[1], W - 1, out=ucol[1])

        # weights
        #   wcol[0, 0] = (1 - uw) * (1 - vw)
        #   wcol[0, 1] = uw * (1 - vw)
        #   wcol[1, 0] = (1 - uw) * vw
        #   wcol[1, 1] = uw * vw
        wcol[0, 1] = uw[i:i_end]
        numpy.subtract(1, wcol[0, 1], out=wcol[0, 0])
        numpy.multiply(wcol[0], vw[i:i_end], out=wcol[1])
        wcol[0] -= wcol[1]

        # packing to the panel whose shape is (B, C, 2, 2, l, out_W)
        panel = x[:, :, vcol[:, None], ucol[None, :]]

        # interpolation
        panel = panel.reshape((B * C, 4, l * out_W))
        weights = wcol.reshape((4, l * out_W))
        iout = i * out_W
        iout_end = i_end * out_W
        numpy.einsum('ijk,jk->ik', panel, weights, out=y[:, iout:iout_end])
        del panel, weights

    return y.reshape((B, C, out_H, out_W))


def interpolate_bilinear_gpu(x, v, u, vw, uw):
    B, C, H, W = x.shape
    out_H, out_W = v.shape
    y = cuda.cupy.empty((B, C, out_H, out_W), dtype=x.dtype)

    cuda.elementwise(
        'raw T x, S v, S u, T vw, T uw, S H, S W, S outsize', 'T y', '''
        // indices
        S v0 = v;
        S v1 = min(v + 1, (S)(H - 1));
        S u0 = u;
        S u1 = min(u + 1, (S)(W - 1));
        // weights
        T w0 = (1 - vw) * (1 - uw);
        T w1 = (1 - vw) * uw;
        T w2 = vw * (1 - uw);
        T w3 = vw * uw;
        // fetch
        S offset = i / outsize * H * W;
        T px0 = x[offset + v0 * W + u0];
        T px1 = x[offset + v0 * W + u1];
        T px2 = x[offset + v1 * W + u0];
        T px3 = x[offset + v1 * W + u1];
        // interpolate
        y = (w0 * px0 + w1 * px1) + (w2 * px2 + w3 * px3);
        ''', 'resize_images_interpolate_bilinear'
    )(x, v, u, vw, uw, H, W, out_H * out_W, y)
    return y


def interpolate_grad_bilinear_cpu(gy, v, u, vw, uw, H, W):
    B, C, out_H, out_W = gy.shape

    # indices
    vcol = numpy.empty((2, out_H, out_W), dtype=v.dtype)
    ucol = numpy.empty((2, out_H, out_W), dtype=u.dtype)
    vcol[0] = v
    ucol[0] = u
    numpy.add(vcol[0], 1, out=vcol[1])
    numpy.add(ucol[0], 1, out=ucol[1])
    numpy.minimum(vcol[1], H - 1, out=vcol[1])
    numpy.minimum(ucol[1], W - 1, out=ucol[1])

    # weights
    wcol = numpy.empty((2, 2, out_H, out_W), dtype=gy.dtype)
    wcol[0, 1] = uw
    numpy.subtract(1, wcol[0, 1], out=wcol[0, 0])
    numpy.multiply(wcol[0], vw, out=wcol[1])
    wcol[0] -= wcol[1]

    # grad
    gycol = gy.reshape((B * C, 1, 1, out_H, out_W)) * wcol

    # ravel everything and use `bincount`
    indices = (vcol[:, None] * W + ucol[None, :]).ravel()
    offsets = numpy.arange(0, B * C * H * W, H * W, dtype=v.dtype)
    indices = (offsets[:, None] + indices).ravel()
    gx = numpy.bincount(indices, weights=gycol.ravel(),
                        minlength=(B * C * H * W))
    gx = gx.astype(gy.dtype, copy=False)

    return gx.reshape((B, C, H, W))


def interpolate_grad_bilinear_gpu(gy, v, u, vw, uw, H, W):
    B, C, out_H, out_W = gy.shape
    gx = cuda.cupy.zeros((B * C, H, W), dtype=gy.dtype)

    cuda.elementwise(
        'T gy, S v, S u, T vw, T uw, S H, S W, S outsize', 'raw T gx', '''
        // indices
        S v0 = v;
        S v1 = min(v + 1, (S)(H - 1));
        S u0 = u;
        S u1 = min(u + 1, (S)(W - 1));
        // weights
        T w0 = (1 - vw) * (1 - uw);
        T w1 = (1 - vw) * uw;
        T w2 = vw * (1 - uw);
        T w3 = vw * uw;
        // scatter
        S offset = i / outsize * H * W;
        atomicAdd(&gx[offset + v0 * W + u0], w0 * gy);
        atomicAdd(&gx[offset + v0 * W + u1], w1 * gy);
        atomicAdd(&gx[offset + v1 * W + u0], w2 * gy);
        atomicAdd(&gx[offset + v1 * W + u1], w3 * gy);
        ''', 'resize_images_interpolate_grad_bilinear'
    )(gy, v, u, vw, uw, H, W, out_H * out_W, gx)

    return gx.reshape((B, C, H, W))


def compute_indices_and_weights(out_size, in_size, mode, align_corners, xp):
    out_H, out_W = out_size
    H, W = in_size
    if mode == 'bilinear':
        if align_corners:
            v = xp.linspace(0, H - 1, num=out_H, dtype=numpy.float)
            u = xp.linspace(0, W - 1, num=out_W, dtype=numpy.float)
        else:
            y_scale = H / out_H
            x_scale = W / out_W
            v = (xp.arange(out_H, dtype=numpy.float) + 0.5) * y_scale - 0.5
            v = xp.maximum(v, 0)
            u = (xp.arange(out_W, dtype=numpy.float) + 0.5) * x_scale - 0.5
            u = xp.maximum(u, 0)
        vw, v = xp.modf(v)
        uw, u = xp.modf(u)
    elif mode == 'nearest':
        y_scale = H / out_H
        x_scale = W / out_W
        v = xp.minimum(xp.floor(
            xp.arange(out_H, dtype=numpy.float) * y_scale), H - 1)
        u = xp.minimum(xp.floor(
            xp.arange(out_W, dtype=numpy.float) * x_scale), W - 1)
        vw = xp.zeros_like(v)
        uw = xp.zeros_like(u)
    return v, u, vw, uw


class ResizeImages(function_node.FunctionNode):

    def __init__(self, output_shape, mode, align_corners):
        self.out_H = output_shape[0]
        self.out_W = output_shape[1]
        assert mode in ['bilinear', 'nearest']
        self.mode = mode
        self.align_corners = align_corners

    def check_type_forward(self, in_types):
        type_check._argname(in_types, ('x',))

        x_type = in_types[0]
        type_check.expect(
            x_type.dtype == numpy.float32,
            x_type.ndim == 4
        )

    def forward(self, inputs):
        x, = inputs
        xp = backend.get_array_module(x)

        v, u, vw, uw = compute_indices_and_weights(
            (self.out_H, self.out_W), x.shape[2:],
            self.mode, self.align_corners, xp)
        v = v.astype(numpy.intp)
        u = u.astype(numpy.intp)
        vw = vw.astype(x.dtype)
        uw = uw.astype(x.dtype)

        # Meshgrid-like operation. Meshgrid can cause
        # performance loss due to memory consumption.
        # Note that numpy 1.9 doesn't support broadcast_to method.
        v, u, vw, uw = xp.broadcast_arrays(
            v[:, None], u[None, :], vw[:, None], uw[None, :])

        if xp is numpy:
            y = interpolate_bilinear_cpu(x, v, u, vw, uw)
        else:
            y = interpolate_bilinear_gpu(x, v, u, vw, uw)
        return y,

    def backward(self, indexes, grad_outputs):
        return ResizeImagesGrad(
            self.inputs[0].shape,
            (self.out_H, self.out_W),
            self.mode, self.align_corners).apply(grad_outputs)


class ResizeImagesGrad(function_node.FunctionNode):

    def __init__(self, input_shape, output_shape, mode, align_corners):
        self.out_H = output_shape[0]
        self.out_W = output_shape[1]
        self.input_shape = input_shape
        assert mode in ['bilinear', 'nearest']
        self.mode = mode
        self.align_corners = align_corners

    def check_type_forward(self, in_types):
        type_check._argname(in_types, ('gy',))

        gy_type = in_types[0]
        type_check.expect(
            gy_type.dtype == numpy.float32,
            gy_type.ndim == 4
        )

    def forward(self, inputs):
        gy, = inputs
        xp = backend.get_array_module(gy)

        _, C, H, W = self.input_shape

        v, u, vw, uw = compute_indices_and_weights(
            (self.out_H, self.out_W), (H, W),
            self.mode, self.align_corners, xp)
        v = v.astype(numpy.intp)
        u = u.astype(numpy.intp)
        vw = vw.astype(gy.dtype)
        uw = uw.astype(gy.dtype)

        # Meshgrid-like operation. Meshgrid can cause
        # performance loss due to memory consumption.
        # Note that numpy 1.9 doesn't support broadcast_to method.
        v, u, vw, uw = xp.broadcast_arrays(
            v[:, None], u[None, :], vw[:, None], uw[None, :])

        if xp is numpy:
            gx = interpolate_grad_bilinear_cpu(gy, v, u, vw, uw, H, W)
        else:
            gx = interpolate_grad_bilinear_gpu(gy, v, u, vw, uw, H, W)
        return gx,

    def backward(self, indexes, grad_outputs):
        return ResizeImages(
            (self.out_H, self.out_W),
            self.mode, self.align_corners).apply(grad_outputs)


def resize_images(x, output_shape, *, mode='bilinear', align_corners=True):
    """Resize images to the given shape.

    This function resizes 2D data to :obj:`output_shape`.

    Notation: here is a notation for dimensionalities.

    - :math:`n` is the batch size.
    - :math:`c_I` is the number of the input channels.
    - :math:`h` and :math:`w` are the height and width of the input image,
      respectively.
    - :math:`h_O` and :math:`w_O` are the height and width of the output
      image.

    Args:
        x (:class:`~chainer.Variable` or :ref:`ndarray`):
            Input variable of shape :math:`(n, c_I, h, w)`.
        output_shape (tuple): This is a tuple of length 2 whose values are
            :obj:`(h_O, w_O)`. Note that the order of height and width is
            opposite of the one in OpenCV.
        mode ({\'bilinear\', \'nearest\'}): Defines the sampling rule.
        align_corners (bool): When this value is :obj:`True`,
            the corners of the input are mapped to the corners of
            the output. When :obj:`False`, the behavior is the same as
            OpenCV.

    Returns:
        ~chainer.Variable: Resized image whose shape is \
            :math:`(n, c_I, h_O, w_O)`.

    """
    return ResizeImages(output_shape, mode, align_corners).apply((x,))[0]



from chainer import backend
from chainer import function_node
from chainer import utils
from chainer.utils import type_check


class Arctanh(function_node.FunctionNode):

    """Elementwise inverse hyperbolic tangent function."""

    def check_type_forward(self, in_types):
        type_check._argname(in_types, ('x',))
        x_type, = in_types

        type_check.expect(x_type.dtype.kind == 'f')

    def forward(self, inputs):
        self.retain_inputs((0,))
        x, = inputs
        xp = backend.get_array_module(x)
        y = xp.arctanh(x)
        return utils.force_array(y, dtype=x.dtype),

    def backward(self, indexes, grad_outputs):
        x, = self.get_retained_inputs()
        gy, = grad_outputs
        gx = 1. / (1 - x ** 2) * gy
        return gx,


def arctanh(x):
    """Elementwise inverse hyperbolic tangent function.

    Args:
        x (:class:`~chainer.Variable` or :ref:`ndarray`): Input variable.

    Returns:
        ~chainer.Variable: Output variable.

    """
    return Arctanh().apply((x,))[0]



from __future__ import absolute_import
import atexit
import functools
import os
import random
import types

import numpy

from chainer.backends import cuda
from chainer.testing import _bundle


_old_python_random_state = None
_old_numpy_random_state = None


def _numpy_do_setup(deterministic=True):
    global _old_python_random_state
    global _old_numpy_random_state
    _old_python_random_state = random.getstate()
    _old_numpy_random_state = numpy.random.get_state()
    if not deterministic:
        numpy.random.seed()
    else:
        numpy.random.seed(100)


def _numpy_do_teardown():
    global _old_python_random_state
    global _old_numpy_random_state
    random.setstate(_old_python_random_state)
    numpy.random.set_state(_old_numpy_random_state)
    _old_python_random_state = None
    _old_numpy_random_state = None


def do_setup(deterministic=True):
    if cuda.available:
        cuda.cupy.testing.random.do_setup(deterministic)
    else:
        _numpy_do_setup(deterministic)


def do_teardown():
    if cuda.available:
        cuda.cupy.testing.random.do_teardown()
    else:
        _numpy_do_teardown()


# In some tests (which utilize condition.repeat or condition.retry),
# setUp/tearDown is nested. _setup_random() and _teardown_random() do their
# work only in the outermost setUp/tearDown pair.
_nest_count = 0


@atexit.register
def _check_teardown():
    assert _nest_count == 0, ('_setup_random() and _teardown_random() '
                              'must be called in pairs.')


def _setup_random():
    """Sets up the deterministic random states of ``numpy`` and ``cupy``.

    """
    global _nest_count
    if _nest_count == 0:
        nondeterministic = bool(int(os.environ.get(
            'CHAINER_TEST_RANDOM_NONDETERMINISTIC', '0')))
        do_setup(not nondeterministic)
    _nest_count += 1


def _teardown_random():
    """Tears down the deterministic random states set up by ``_setup_random``.

    """
    global _nest_count
    assert _nest_count > 0, '_setup_random has not been called'
    _nest_count -= 1
    if _nest_count == 0:
        do_teardown()


def generate_seed():
    assert _nest_count > 0, 'random is not set up'
    return numpy.random.randint(0xffffffff)


def _fix_random(setup_method_name, teardown_method_name):
    # TODO(niboshi): Prevent this decorator from being applied within
    #    condition.repeat or condition.retry decorators. That would repeat
    #    tests with the same random seeds. It's okay to apply this outside
    #    these decorators.

    def decorator(impl):
        if (isinstance(impl, types.FunctionType) and
                impl.__name__.startswith('test_')):
            # Applied to test method
            @functools.wraps(impl)
            def test_func(self, *args, **kw):
                _setup_random()
                try:
                    impl(self, *args, **kw)
                finally:
                    _teardown_random()
            return test_func

        if isinstance(impl, _bundle._ParameterizedTestCaseBundle):
            cases = impl
        else:
            tup = _bundle._TestCaseTuple(impl, None, None)
            cases = _bundle._ParameterizedTestCaseBundle([tup])

        for klass, _, _ in cases.cases:
            # Applied to test case class

            def make_methods():
                # make_methods is required to bind the variables prev_setup and
                # prev_teardown.
                prev_setup = getattr(klass, setup_method_name)
                prev_teardown = getattr(klass, teardown_method_name)

                @functools.wraps(prev_setup)
                def new_setup(self):
                    _setup_random()
                    prev_setup(self)

                @functools.wraps(prev_teardown)
                def new_teardown(self):
                    try:
                        prev_teardown(self)
                    finally:
                        _teardown_random()

                return new_setup, new_teardown

            setup, teardown = make_methods()

            setattr(klass, setup_method_name, setup)
            setattr(klass, teardown_method_name, teardown)

        return cases

    return decorator


def fix_random():
    """Decorator that fixes random numbers in a test.

    This decorator can be applied to either a test case class or a test method.
    It should not be applied within ``condition.retry`` or
    ``condition.repeat``.
    """
    return _fix_random('setUp', 'tearDown')



import inspect

from chainerx import _core
from chainerx._docs import array
from chainerx._docs import backend
from chainerx._docs import backprop
from chainerx._docs import context
from chainerx._docs import device
from chainerx._docs import routines
from chainerx._docs import utils


def set_doc(obj, docstring):
    if ((inspect.ismethoddescriptor(obj) or inspect.isroutine(obj))
            and not inspect.isfunction(obj)):
        # pybind-generated functions and methods
        _core._set_pybind_doc(obj, docstring)
        return

    obj.__doc__ = docstring


def set_docs():
    for m in (array, backend, backprop, context, device, routines, utils):
        m.set_docs()



import glob
import os
import warnings

import numpy as np
import onnx

try:
    import onnxruntime as rt
    ONNXRUNTIME_AVAILABLE = True
except ImportError:
    warnings.warn(
        'ONNXRuntime is not installed. Please install it to use '
        ' the testing utility for ONNX-Chainer\'s converters.',
        ImportWarning)
    ONNXRUNTIME_AVAILABLE = False


def load_test_data(data_dir, input_names, output_names):
    inout_values = []
    for kind, names in [('input', input_names), ('output', output_names)]:
        names = list(names)
        values = {}
        for pb in sorted(
                glob.glob(os.path.join(data_dir, '{}_*.pb'.format(kind)))):
            tensor = onnx.load_tensor(pb)
            if tensor.name in names:
                name = tensor.name
                names.remove(name)
            else:
                name = names.pop(0)
            values[name] = onnx.numpy_helper.to_array(tensor)
        inout_values.append(values)
    return tuple(inout_values)


def check_model_expect(test_path, input_names=None, rtol=1e-5, atol=1e-5):
    if not ONNXRUNTIME_AVAILABLE:
        raise ImportError('ONNX Runtime is not found on checking module.')

    model_path = os.path.join(test_path, 'model.onnx')
    with open(model_path, 'rb') as f:
        onnx_model = onnx.load_model(f)
    sess = rt.InferenceSession(onnx_model.SerializeToString())
    rt_input_names = [value.name for value in sess.get_inputs()]
    rt_output_names = [value.name for value in sess.get_outputs()]

    # To detect unexpected inputs created by exporter, check input names
    if input_names is not None:
        assert list(sorted(input_names)) == list(sorted(rt_input_names))

    test_data_sets = sorted([
        p for p in os.listdir(test_path) if p.startswith('test_data_set_')])
    for test_data in test_data_sets:
        test_data_path = os.path.join(test_path, test_data)
        assert os.path.isdir(test_data_path)
        inputs, outputs = load_test_data(
            test_data_path, rt_input_names, rt_output_names)

        rt_out = sess.run(list(outputs.keys()), inputs)
        for cy, my in zip(outputs.values(), rt_out):
            np.testing.assert_allclose(cy, my, rtol=rtol, atol=atol)



import ctypes

import mpi4py.MPI
import numpy as np
from chainermn.communicators import _communication_utility

import chainer.backends
import chainerx as chx

try:
    import cupy as cp
    _cupy_avail = True
except Exception:
    cp = None
    _cupy_avail = False


def _get_memory_pointer_from_chainerx(array):
    # Currently, ChainerMN requires CuPy to support ChainerX.
    # This is because ChainerX's backend does not provide a raw
    # memory pointer class.
    return cp.cuda.MemoryPointer(
        cp.cuda.UnownedMemory(
            array.data_ptr + array.offset,
            array.data_size,
            array,
            array.device.index),
        0)


class ParamsData(object):
    def __init__(self, params, attr_name, zero_fill):
        n_params = len(params)
        params_dptr = np.empty(n_params, dtype=np.int64)
        params_dtype = np.empty(n_params, dtype=np.int32)
        params_size_csum = np.empty(n_params+1, dtype=np.int32)
        params_size_csum[0] = 0
        for i, param in enumerate(params):
            v = getattr(param, attr_name)
            if attr_name == 'grad' and v is None and zero_fill:
                v = param.xp.zeros_like(param.data)
                setattr(param, attr_name, v)
            xp = chainer.backend.get_array_module(v)

            if xp == cp:
                v_data = v.data
            elif xp == chx:
                v_data = _get_memory_pointer_from_chainerx(v)
            else:
                raise ValueError(
                    '{} is from an unsupported array module'.format(type(v)))

            params_dptr[i] = v_data.ptr
            if v.dtype not in [np.float16, np.float32, np.float64]:
                raise ValueError('dtype must be float16, float32 or float64.')
            params_dtype[i] = _communication_utility._get_nccl_type_id(v.dtype)
            params_size_csum[i+1] = params_size_csum[i] + v.size
        self.n_params = n_params
        self.n_elems = params_size_csum[n_params]
        self.size_csum = chainer.cuda.cupy.asarray(params_size_csum)
        self.dtype = chainer.cuda.cupy.asarray(params_dtype)
        self.dptr = chainer.cuda.cupy.asarray(params_dptr)


class HostPinnedMemory(object):

    def __init__(self):
        if not _cupy_avail:
            raise RuntimeError('HostPinnedMemory cannot be used: ' +
                               'Cupy is not available.')
        self.size = 0
        self.memory = None

    def assign(self, size):
        if size > self.size:
            self.size = size
            self.memory = cp.cuda.alloc_pinned_memory(size)

    def ptr(self, offset=0):
        return ctypes.c_void_p(self.memory.ptr + offset)

    def buffer(self, size):
        return ctypes.cast(
            self.memory.ptr,
            ctypes.POINTER(ctypes.c_ubyte * size)
        ).contents

    def array(self, count, offset=0, dtype=np.float32):
        if dtype is None:
            raise TypeError('dtype must be an instance of numpy.dtype class')
        return np.frombuffer(
            self.memory, count=count, offset=offset, dtype=dtype)


class DeviceMemory(object):

    def __init__(self):
        if not _cupy_avail:
            raise RuntimeError('DeviceMemory cannot be used: ' +
                               'Cupy is not available.')
        self.size = 0
        self.memory = None

    def assign(self, size):
        if size > self.size:
            self.size = size
            self.memory = cp.cuda.alloc(size)

    def from_device(self, src, size, offset=0, stream=None):
        dst = self.memory + offset
        xp = chainer.backend.get_array_module(src)
        if xp == cp:
            src_data = src.data
        elif xp == chx:
            src_data = _get_memory_pointer_from_chainerx(src)
        else:
            raise ValueError(
                '{} is from an unsupported array module'.format(type(src)))
        if stream is None:
            dst.copy_from_device(src_data, size)
        else:
            dst.copy_from_device_async(src_data, size, stream)

    def to_device(self, dst, size, offset=0, stream=None):
        src = self.memory + offset
        xp = chainer.backend.get_array_module(dst)
        if xp == cp:
            dst_data = dst.data
        elif xp == chx:
            dst_data = _get_memory_pointer_from_chainerx(dst)
        else:
            raise ValueError(
                '{} is from an unsupported array module'.format(type(dst)))
        if stream is None:
            dst_data.copy_from_device(src, size)
        else:
            dst_data.copy_from_device_async(src, size, stream)

    def ptr(self):
        return self.memory.ptr

    def buffer(self, size):
        return ctypes.cast(
            self.memory.ptr,
            ctypes.POINTER(ctypes.c_ubyte * size)
        ).contents

    def array(self, shape, offset=0, dtype=np.float32):
        if dtype is None:
            raise TypeError('dtype must be an instance of numpy.dtype class')
        return cp.ndarray(shape, memptr=self.memory + offset, dtype=dtype)


def extract_params_set_data(model):
    return [param for _, param in sorted(model.namedparams())
            if param.data is not None]


def extract_params_set_grad(model, zero_fill):
    if zero_fill:
        return [param for _, param in sorted(model.namedparams())
                if param.data is not None]
    else:
        return [param for _, param in sorted(model.namedparams())
                if param.data is not None and param.grad is not None]


def count_grad_elements(params, zero_fill):
    if zero_fill:
        return sum(param.data.size for param in params)
    else:
        return sum(param.grad.size for param in params)


def pack_params(params, attr_name, buffer,
                transfer_dtype, zero_fill, stream=None):
    if len(params) == 0:
        return

    # NOTE: dtypes of params might be mixed, in particular f16 & f32.
    offset = 0
    for param in params:
        v = getattr(param, attr_name)
        if attr_name == 'grad' and v is None and zero_fill:
            v = param.xp.zeros_like(param.data)
        size = v.size * np.dtype(transfer_dtype).itemsize
        if v.dtype != transfer_dtype:
            tmp = v.astype(transfer_dtype)
            buffer.from_device(tmp, size, offset, stream)
        else:
            buffer.from_device(v, size, offset, stream)

        offset += size


def unpack_params(params, attr_name, buffer,
                  transfer_dtype, zero_fill, stream=None):
    """Pack parameters into a single CuPy array for efficient communication."""
    if len(params) == 0:
        return
    xp = chainer.backend.get_array_module(getattr(params[0], attr_name))
    offset = 0
    for param in params:
        v = getattr(param, attr_name)
        if attr_name == 'grad' and v is None and zero_fill:
            v = param.xp.empty_like(param.data)
            setattr(param, attr_name, v)
        size = v.size * np.dtype(transfer_dtype).itemsize
        grad_dtype = v.dtype
        if grad_dtype != transfer_dtype:
            v = xp.array(v, copy=False, dtype=transfer_dtype)
        buffer.to_device(v, size, offset, stream)
        offset += size
        if grad_dtype != transfer_dtype:
            # avoid using setattr as ChainerX array cannot be directly updated
            getattr(param, attr_name)[...] = v.astype(grad_dtype)


def array_to_buffer_object(array, mpi_dtype=mpi4py.MPI.FLOAT):
    xp = chainer.backend.get_array_module(array)

    if xp is np:
        return get_device_memory_pointer(array)
    else:
        return (get_device_memory_pointer(array), mpi_dtype)


def get_device_memory_pointer(array):
    xp = chainer.backend.get_array_module(array)
    array = xp.ascontiguousarray(array)

    if xp is np:
        return array
    elif xp is cp:
        return ctypes.cast(
            array.data.ptr,
            ctypes.POINTER(ctypes.c_ubyte * array.nbytes)
        ).contents
    elif xp is chx:
        backend_name = array.device.backend.name
        if backend_name not in ['native', 'cuda']:
            raise ValueError(
                '{} is an unsupported backend'.format(backend_name))
        return ctypes.cast(
            array.data_ptr,
            ctypes.POINTER(ctypes.c_ubyte * array.nbytes)
        ).contents
    else:
        raise ValueError(
            '{} is from an unsupported array module'.format(type(array)))


def _batched_pack_params(params_data, buffer, dtype, stream=None):
    n_params = params_data.n_params
    n_elems = params_data.n_elems
    params_dptr = params_data.dptr
    params_dtype = params_data.dtype
    params_size_csum = params_data.size_csum
    buf_dtype = _communication_utility._get_nccl_type_id(dtype)
    n_threads = 128
    n_blocks = (n_elems + n_threads - 1) // n_threads
    if stream is None:
        stream = cp.cuda.get_current_stream()
    with stream:
        _cupy_batched_pack_params()(
            (n_blocks, ), (n_threads, ),
            (buffer.memory.ptr, buf_dtype, n_elems,
             params_dptr, params_dtype, params_size_csum, n_params))


def _batched_unpack_params(params_data, buffer, dtype, stream=None):
    n_params = params_data.n_params
    n_elems = params_data.n_elems
    params_dptr = params_data.dptr
    params_dtype = params_data.dtype
    params_size_csum = params_data.size_csum
    buf_dtype = _communication_utility._get_nccl_type_id(dtype)
    n_threads = 128
    n_blocks = (n_elems + n_threads - 1) // n_threads
    if stream is None:
        stream = cp.cuda.get_current_stream()
    with stream:
        _cupy_batched_unpack_params()(
            (n_blocks, ), (n_threads, ),
            (buffer.memory.ptr, buf_dtype, n_elems,
             params_dptr, params_dtype, params_size_csum, n_params))


def _cupy_batched_pack_params():
    return chainer.cuda.raw(r'''
#include <cupy/carray.cuh>
#define NCCL_FLOAT16  6
#define NCCL_FLOAT32  7
#define NCCL_FLOAT64  8
    extern "C" __global__
    void cupy_batched_pack_params(
            void *dst0, int dst_dtype, int n_elems,
            unsigned long *params_dptr, int *params_dtype,
            int *params_size_csum, int n_params) {
        int tid = threadIdx.x + blockIdx.x * blockDim.x;
        if (tid >= n_elems) return;
        int j_min = 0;
        int j_max = n_params - 1;
        int j;
        while (1) {
            j = (j_min + j_max) / 2;
            if (tid < params_size_csum[j]) {
                j_max = j - 1;
                continue;
            }
            if (tid >= params_size_csum[j+1]){
                j_min = j + 1;
                continue;
            }
            break;
        }
        assert(tid >= params_size_csum[j]);
        assert(tid < params_size_csum[j+1]);
        int src_dtype = params_dtype[j];
        int src_idx = tid - params_size_csum[j];
        if (dst_dtype == NCCL_FLOAT16) {
            half* dst = (half*) dst0;
            if (src_dtype == NCCL_FLOAT16) {
                dst[tid] = (half) (((half*) (params_dptr[j]))[src_idx]);
            }
            else if (src_dtype == NCCL_FLOAT32) {
                dst[tid] = (half) (((float*) (params_dptr[j]))[src_idx]);
            }
            else if (src_dtype == NCCL_FLOAT64) {
                dst[tid] = (half) (((double*) (params_dptr[j]))[src_idx]);
            }
        }
        else if (dst_dtype == NCCL_FLOAT32) {
            float* dst = (float*) dst0;
            if (src_dtype == NCCL_FLOAT16) {
                dst[tid] = (float) (((half*) (params_dptr[j]))[src_idx]);
            }
            else if (src_dtype == NCCL_FLOAT32) {
                dst[tid] = (float) (((float*) (params_dptr[j]))[src_idx]);
            }
            else if (src_dtype == NCCL_FLOAT64) {
                dst[tid] = (float) (((double*) (params_dptr[j]))[src_idx]);
            }
       }
       else if (dst_dtype == NCCL_FLOAT64) {
            double* dst = (double*) dst0;
            if (src_dtype == NCCL_FLOAT16) {
                dst[tid] = (double) (((half*) (params_dptr[j]))[src_idx]);
            }
            else if (src_dtype == NCCL_FLOAT32) {
                dst[tid] = (double) (((float*) (params_dptr[j]))[src_idx]);
            }
            else if (src_dtype == NCCL_FLOAT64) {
                dst[tid] = (double) (((double*) (params_dptr[j]))[src_idx]);
            }
       }
    }
    ''', 'cupy_batched_pack_params')


def _cupy_batched_unpack_params():
    return chainer.cuda.raw(r'''
#include <cupy/carray.cuh>
#define NCCL_FLOAT16  6
#define NCCL_FLOAT32  7
#define NCCL_FLOAT64  8
    extern "C" __global__
    void cupy_batched_unpack_params(
            void *src0, int src_dtype, int n_elems,
            unsigned long *params_dptr, int *params_dtype,
            int *params_size_csum, int n_params) {
        int tid = threadIdx.x + blockIdx.x * blockDim.x;
        if (tid >= n_elems) return;
        int j_min = 0;
        int j_max = n_params - 1;
        int j;
        while (1) {
            j = (j_min + j_max) / 2;
            if (tid < params_size_csum[j]) {
                j_max = j - 1;
                continue;
            }
            if (tid >= params_size_csum[j+1]){
                j_min = j + 1;
                continue;
            }
            break;
        }
        assert(tid >= params_size_csum[j]);
        assert(tid < params_size_csum[j+1]);
        int dst_dtype = params_dtype[j];
        int dst_idx = tid - params_size_csum[j];
        if (src_dtype == NCCL_FLOAT16) {
            half* src = (half*) src0;
            if (dst_dtype == NCCL_FLOAT16) {
                ((half*) (params_dptr[j]))[dst_idx] = (half) src[tid];
            }
            else if (dst_dtype == NCCL_FLOAT32) {
                ((float*) (params_dptr[j]))[dst_idx] = (float) src[tid];
            }
            else if (dst_dtype == NCCL_FLOAT64) {
                ((double*) (params_dptr[j]))[dst_idx] = (double) src[tid];
            }
        }
        else if (src_dtype == NCCL_FLOAT32) {
            float* src = (float*) src0;
            if (dst_dtype == NCCL_FLOAT16) {
                ((half*) (params_dptr[j]))[dst_idx] = (half) src[tid];
            }
            else if (dst_dtype == NCCL_FLOAT32) {
                ((float*) (params_dptr[j]))[dst_idx] = (float) src[tid];
            }
            else if (dst_dtype == NCCL_FLOAT64) {
                ((double*) (params_dptr[j]))[dst_idx] = (double) src[tid];
            }
       }
       else if (src_dtype == NCCL_FLOAT64) {
            double* src = (double*) src0;
            if (dst_dtype == NCCL_FLOAT16) {
                ((half*) (params_dptr[j]))[dst_idx] = (half) src[tid];
            }
            else if (dst_dtype == NCCL_FLOAT32) {
                ((float*) (params_dptr[j]))[dst_idx] = (float) src[tid];
            }
            else if (dst_dtype == NCCL_FLOAT64) {
                ((double*) (params_dptr[j]))[dst_idx] = (double) src[tid];
            }
       }
    }''', 'cupy_batched_unpack_params')



import chainer
from chainer.backends import cuda
from chainer.functions.normalization import batch_normalization
import chainer.utils


class _MpiImpl(batch_normalization.GeneralBatchNormalizationImpl):
    def __init__(self, comm):
        self.comm = comm

    def get_mean_and_var(self, axis, gamma, x, xp, interm_dtype):
        tmp = xp.empty(gamma.size * 2, dtype=gamma.dtype)
        x.mean(axis=axis, out=tmp[:gamma.size], dtype=gamma.dtype)
        xp.square(x).mean(axis=axis, out=tmp[gamma.size:], dtype=gamma.dtype)
        if xp is cuda.cupy:
            chainer.cuda.Stream.null.synchronize()
        self.comm._multi_node_mean(None, tmp)
        mean = tmp[:gamma.size]
        sqmean = tmp[gamma.size:]
        var = sqmean - xp.square(mean)
        return mean, var

    def get_ggamma_and_gbeta(self, axis, gamma, gy, x_hat, xp):
        tmp = xp.empty(gamma.size * 2, dtype=gamma.dtype)
        gy.sum(axis=axis, out=tmp[:gamma.size], dtype=gamma.dtype)
        (gy * x_hat).sum(axis=axis, out=tmp[gamma.size:], dtype=gamma.dtype)
        if xp is cuda.cupy:
            chainer.cuda.Stream.null.synchronize()
        self.comm._multi_node_mean(None, tmp)
        gbeta = tmp[:gamma.size]
        ggamma = tmp[gamma.size:]
        return gbeta, ggamma


class _NcclImpl(batch_normalization.GeneralBatchNormalizationImpl):

    def __init__(self, comm):
        self.comm = comm

        # We need to delay importing MPI4py (and momdules that import MPI4py)
        import chainermn.communicators._memory_utility as memory_utility_module
        self.memory_utility_module = memory_utility_module

    def get_mean_and_var(self, axis, gamma, x, xp, interm_dtype):
        gpu_buffer_n_elems = gamma.size * 2
        gpu_buffer_size = gamma.dtype.itemsize * gpu_buffer_n_elems
        gpu_buffer_a = self.memory_utility_module.DeviceMemory()
        gpu_buffer_b = self.memory_utility_module.DeviceMemory()
        gpu_buffer_a.assign(gpu_buffer_size)
        gpu_buffer_b.assign(gpu_buffer_size)
        gpu_buffer_a_array = gpu_buffer_a.array(
            gpu_buffer_n_elems, dtype=gamma.dtype)
        x.mean(axis=axis, out=gpu_buffer_a_array[:gamma.size],
               dtype=gamma.dtype)
        xp.square(x).mean(axis=axis, out=gpu_buffer_a_array[gamma.size:],
                          dtype=gamma.dtype)
        self.comm._multi_node_mean_nccl(gpu_buffer_a,
                                        gpu_buffer_b,
                                        gpu_buffer_n_elems,
                                        gamma.dtype)
        gpu_buffer_a_array = gpu_buffer_b.array(
            gpu_buffer_n_elems,
            dtype=gamma.dtype)

        mean = gpu_buffer_a_array[:gamma.size]
        sqmean = gpu_buffer_a_array[gamma.size:]
        var = sqmean - xp.square(mean)
        return mean, var

    def get_ggamma_and_gbeta(self, axis, gamma, gy, x_hat, xp):
        gpu_buffer_n_elems = gamma.size * 2
        gpu_buffer_size = gamma.dtype.itemsize * gpu_buffer_n_elems
        gpu_buffer_a = self.memory_utility_module.DeviceMemory()
        gpu_buffer_b = self.memory_utility_module.DeviceMemory()
        gpu_buffer_a.assign(gpu_buffer_size)
        gpu_buffer_b.assign(gpu_buffer_size)
        gpu_buffer_a_array = gpu_buffer_a.array(
            gpu_buffer_n_elems, dtype=gamma.dtype)
        gy.sum(axis=axis, out=gpu_buffer_a_array[:gamma.size],
               dtype=gamma.dtype)
        (gy * x_hat).sum(axis=axis, out=gpu_buffer_a_array[gamma.size:],
                         dtype=gamma.dtype)
        self.comm._multi_node_mean_nccl(gpu_buffer_a,
                                        gpu_buffer_b,
                                        gpu_buffer_n_elems,
                                        gamma.dtype)
        gpu_buffer_a_array = gpu_buffer_b.array(
            gpu_buffer_n_elems,
            dtype=gamma.dtype)

        gbeta = gpu_buffer_a_array[:gamma.size]
        ggamma = gpu_buffer_a_array[gamma.size:]
        return gbeta, ggamma


def get_communication_backend(comm, communication_backend='auto'):
    if communication_backend not in ['mpi', 'nccl', 'auto']:
        raise ValueError('MultiNodeBatchNormalization does not support '
                         '{}.'.format(communication_backend))
    from chainermn.communicators.pure_nccl_communicator \
        import PureNcclCommunicator
    if communication_backend != 'auto':
        if 'nccl' == communication_backend:
            if not isinstance(comm, PureNcclCommunicator):
                raise ValueError('{} is not supported in '
                                 'MultiNodeBatchNormalization when using '
                                 '{}.'.format(communication_backend,
                                              type(comm)))
        selected_communication_backend = communication_backend
    else:
        if isinstance(comm, PureNcclCommunicator):
            selected_communication_backend = 'nccl'
        else:
            selected_communication_backend = 'mpi'
    return selected_communication_backend


class MultiNodeBNImplSelector:
    def __init__(self, comm, communication_backend_name):
        self.comm = comm
        self.communication_backend_name = communication_backend_name

    def __call__(self, batch_norm_func, inputs):
        if self.communication_backend_name == 'nccl':
            return _NcclImpl(self.comm)
        else:
            return _MpiImpl(self.comm)



from chainer.functions.activation import sigmoid
from chainer.functions.activation import tanh
from chainer.functions.math import linear_interpolate
from chainer import link
from chainer.links.connection import linear
from chainer import variable


class GRUBase(link.Chain):

    def __init__(self, in_size, out_size, init=None,
                 inner_init=None, bias_init=None):
        super(GRUBase, self).__init__()
        with self.init_scope():
            self.W_r = linear.Linear(
                in_size, out_size, initialW=init, initial_bias=bias_init)
            self.U_r = linear.Linear(
                out_size, out_size, initialW=inner_init,
                initial_bias=bias_init)
            self.W_z = linear.Linear(
                in_size, out_size, initialW=init, initial_bias=bias_init)
            self.U_z = linear.Linear(
                out_size, out_size, initialW=inner_init,
                initial_bias=bias_init)
            self.W = linear.Linear(
                in_size, out_size, initialW=init, initial_bias=bias_init)
            self.U = linear.Linear(
                out_size, out_size, initialW=inner_init,
                initial_bias=bias_init)


class StatelessGRU(GRUBase):

    """Stateless Gated Recurrent Unit function (GRU).

    GRU function has six parameters :math:`W_r`, :math:`W_z`, :math:`W`,
    :math:`U_r`, :math:`U_z`, and :math:`U`.
    The three parameters :math:`W_r`, :math:`W_z`, and :math:`W` are
    :math:`n \\times m` matrices, and the others :math:`U_r`, :math:`U_z`,
    and :math:`U` are :math:`n \\times n` matrices, where :math:`m` is the
    length of input vectors and :math:`n` is the length of hidden vectors.

    Given two inputs a previous hidden vector :math:`h` and an input vector
    :math:`x`, GRU returns the next hidden vector :math:`h'` defined as

    .. math::

       r &=& \\sigma(W_r x + U_r h), \\\\
       z &=& \\sigma(W_z x + U_z h), \\\\
       \\bar{h} &=& \\tanh(W x + U (r \\odot h)), \\\\
       h' &=& (1 - z) \\odot h + z \\odot \\bar{h},

    where :math:`\\sigma` is the sigmoid function, and :math:`\\odot` is the
    element-wise product.

    As the name indicates, :class:`~chainer.links.StatelessGRU` is *stateless*,
    meaning that it does not hold the value of
    hidden vector :math:`h`.
    For a *stateful* GRU, use :class:`~chainer.links.StatefulGRU`.

    Args:
        in_size(int): Dimension of input vector :math:`x`.
            If ``None``, parameter initialization will be deferred
            until the first forward data pass
            at which time the size will be determined.
        out_size(int): Dimension of hidden vector :math:`h`,
            :math:`\\bar{h}` and :math:`h'`.

    See:
        - `On the Properties of Neural Machine Translation: Encoder-Decoder
          Approaches <https://www.aclweb.org/anthology/W14-4012>`_
          [Cho+, SSST2014].
        - `Empirical Evaluation of Gated Recurrent Neural Networks on Sequence
          Modeling <https://arxiv.org/abs/1412.3555>`_
          [Chung+NIPS2014 DLWorkshop].


    .. seealso:: :class:`~chainer.links.StatefulGRU`

    .. admonition:: Example

        There are several ways to make a ``StatelessGRU`` link.
        Let ``x`` be a two-dimensional input array:

        >>> in_size = 10
        >>> out_size = 20
        >>> x = np.zeros((1, in_size), dtype=np.float32)
        >>> h = np.zeros((1, out_size), dtype=np.float32)

        1. Give both  ``in_size`` and ``out_size`` arguments:

            >>> l = L.StatelessGRU(in_size, out_size)
            >>> h_new = l(h, x)
            >>> h_new.shape
            (1, 20)

        2. Omit ``in_size`` argument or fill it with ``None``:

            >>> l = L.StatelessGRU(None, out_size)
            >>> h_new = l(h, x)
            >>> h_new.shape
            (1, 20)

    """

    def forward(self, h, x):
        r = sigmoid.sigmoid(self.W_r(x) + self.U_r(h))
        z = sigmoid.sigmoid(self.W_z(x) + self.U_z(h))
        h_bar = tanh.tanh(self.W(x) + self.U(r * h))
        h_new = linear_interpolate.linear_interpolate(z, h_bar, h)
        return h_new


class StatefulGRU(GRUBase):
    """Stateful Gated Recurrent Unit function (GRU).

    Stateful GRU function has six parameters :math:`W_r`, :math:`W_z`,
    :math:`W`, :math:`U_r`, :math:`U_z`, and :math:`U`.
    The three parameters :math:`W_r`, :math:`W_z`, and :math:`W` are
    :math:`n \\times m` matrices, and the others :math:`U_r`, :math:`U_z`,
    and :math:`U` are :math:`n \\times n` matrices, where :math:`m` is the
    length of input vectors and :math:`n` is the length of hidden vectors.

    Given input vector :math:`x`, Stateful GRU returns the next
    hidden vector :math:`h'` defined as

    .. math::

       r &=& \\sigma(W_r x + U_r h), \\\\
       z &=& \\sigma(W_z x + U_z h), \\\\
       \\bar{h} &=& \\tanh(W x + U (r \\odot h)), \\\\
       h' &=& (1 - z) \\odot h + z \\odot \\bar{h},

    where :math:`h` is current hidden vector.

    As the name indicates, :class:`~chainer.links.StatefulGRU` is *stateful*,
    meaning that it also holds the next hidden vector `h'` as a state.
    For a *stateless* GRU, use :class:`~chainer.links.StatelessGRU`.

    Args:
        in_size(int): Dimension of input vector :math:`x`.
        out_size(int): Dimension of hidden vector :math:`h`.
        init: Initializer for GRU's input units (:math:`W`).
            It is a callable that takes :ref:`ndarray` and edits its value.
            If it is ``None``, the default initializer is used.
        inner_init: Initializer for the GRU's inner
            recurrent units (:math:`U`).
            It is a callable that takes :ref:`ndarray` and edits its value.
            If it is ``None``, the default initializer is used.
        bias_init: Bias initializer.
            It is a callable that takes :ref:`ndarray` and edits its value.
            If ``None``, the bias is set to zero.

    Attributes:
        h(~chainer.Variable): Hidden vector that indicates the state of
            :class:`~chainer.links.StatefulGRU`.

    .. seealso::
        * :class:`~chainer.links.StatelessGRU`
        * :class:`~chainer.links.GRU`: an alias of
          :class:`~chainer.links.StatefulGRU`

    .. admonition:: Example

        There are several ways to make a ``StatefulGRU`` link.
        Let ``x`` be a two-dimensional input array:

        >>> in_size = 10
        >>> out_size = 20
        >>> x = np.zeros((1, in_size), dtype=np.float32)

        1. Give only ``in_size`` and ``out_size`` arguments:

            >>> l = L.StatefulGRU(in_size, out_size)
            >>> h_new = l(x)
            >>> h_new.shape
            (1, 20)

        2. Give all optional arguments:

            >>> init = np.zeros((out_size, in_size), dtype=np.float32)
            >>> inner_init = np.zeros((out_size, out_size), dtype=np.float32)
            >>> bias = np.zeros((1, out_size), dtype=np.float32)
            >>> l = L.StatefulGRU(in_size, out_size, init=init,
            ...     inner_init=inner_init, bias_init=bias)
            >>> h_new = l(x)
            >>> h_new.shape
            (1, 20)

    """

    def __init__(self, in_size, out_size, init=None,
                 inner_init=None, bias_init=0):
        super(StatefulGRU, self).__init__(
            in_size, out_size, init, inner_init, bias_init)
        self.state_size = out_size
        self.reset_state()

    def device_resident_accept(self, visitor):
        super(StatefulGRU, self).device_resident_accept(visitor)
        if self.h is not None:
            visitor.visit_variable(self.h)

    def set_state(self, h):
        assert isinstance(h, variable.Variable)
        h.to_device(self.device)
        self.h = h

    def reset_state(self):
        self.h = None

    def forward(self, x):
        z = self.W_z(x)
        h_bar = self.W(x)
        if self.h is not None:
            r = sigmoid.sigmoid(self.W_r(x) + self.U_r(self.h))
            z += self.U_z(self.h)
            h_bar += self.U(r * self.h)
        z = sigmoid.sigmoid(z)
        h_bar = tanh.tanh(h_bar)

        if self.h is not None:
            h_new = linear_interpolate.linear_interpolate(z, h_bar, self.h)
        else:
            h_new = z * h_bar
        self.h = h_new
        return self.h


class GRU(StatefulGRU):
    """Stateful Gated Recurrent Unit function (GRU)

    This is an alias of :class:`~chainer.links.StatefulGRU`.

    """

    def forward(self, *args):
        """forward(self, x)

        Does forward propagation.

        """

        n_args = len(args)
        msg = ('Invalid argument. The length of GRU.forward must be 1. '
               'But %d is given. ' % n_args)

        if n_args == 0 or n_args >= 3:
            raise ValueError(msg)
        elif n_args == 2:
            msg += ('In Chainer v2, chainer.links.GRU is changed '
                    'from stateless to stateful. '
                    'One possibility is you assume GRU to be stateless. '
                    'Use chainer.links.StatelessGRU instead.')
            raise ValueError(msg)

        return super(GRU, self).forward(args[0])



import os

import onnx_chainer


TEST_OUT_DIR = 'out'


def gen_test_data_set(model, args, name, opset_version, **kwargs):
    model.xp.random.seed(42)
    test_path = os.path.join(
        TEST_OUT_DIR, 'opset{}'.format(opset_version), name)
    onnx_chainer.export_testcase(
        model, args, test_path, opset_version=opset_version, **kwargs)
    return test_path



from chainer.dataset.tabular import tabular_dataset


class DelegateDataset(tabular_dataset.TabularDataset):

    """A helper class to implement a TabularDataset.

    This class wraps an instance of :class:`~chainer.dataset.TabularDataset`
    and provides methods of :class:`~chainer.dataset.TabularDataset`.
    This class is useful to create a custom dataset class by inheriting it.

    >>> import numpy as np
    >>>
    >>> from chainer.dataset import tabular
    >>>
    >>> class MyDataset(tabular.DelegateDataset):
    ...
    ...     def __init__(self):
    ...         super().__init__(tabular.from_data((
    ...             ('a', np.arange(10)),
    ...             ('b', self.get_b),
    ...             ('c', [3, 1, 4, 5, 9, 2, 6, 8, 7, 0]),
    ...             (('d', 'e'), self.get_de))))
    ...
    ...     def get_b(self, i):
    ...         return 'b[{}]'.format(i)
    ...
    ...     def get_de(self, i):
    ...         return {'d': 'd[{}]'.format(i), 'e': 'e[{}]'.format(i)}
    ...
    >>> dataset = MyDataset()
    >>> len(dataset)
    10
    >>> dataset.keys
    ('a', 'b', 'c', 'd', 'e')
    >>> dataset[0]
    (0, 'b[0]', 3, 'd[0]', 'e[0]')

    Args:
        dataset (chainer.dataset.TabularDataset): An underlying dataset.

    """

    def __init__(self, dataset):
        self.dataset = dataset

    def __len__(self):
        return len(self.dataset)

    @property
    def keys(self):
        return self.dataset.keys

    @property
    def mode(self):
        return self.dataset.mode

    def get_examples(self, indices, key_indices):
        return self.dataset.get_examples(indices, key_indices)



from __future__ import division
from chainer.training import extension, util
from chainer import Variable


class ObservationAggregator(extension.Extension):

    """Trainer extension to aggregate an observation in the trainer.
    Args:
         comm: ChainerMN communicator
         original_key (str): Key of the observation to be summarized.
         If the observation is a :class:`chainer.Variable`, its value
         is automatically copied to CPU.
         aggregated_key (str): Name of the key after the summarization.
         If not specified, it is set to `original_key` to overwrite it.
         comm_trigger: Trigger that decides the timing to communicate
         observation values for aggregation.
         aggregator (function): Function to compute summarization from
         individual values. It takes a list of lists of observed values.
         Each list contains all the observed values since
         the last communication.
    """

    trigger = 1, 'iteration'
    priority = extension.PRIORITY_EDITOR
    name = None

    def __init__(self, comm, original_key, aggregated_key=None,
                 *, comm_trigger=(1, 'iteration'), aggregator=None):
        self.comm = comm
        self.original_key = original_key

        if aggregated_key is None:
            self.aggregated_key = original_key
        else:
            self.aggregated_key = aggregated_key

        self.comm_trigger = util.get_trigger(comm_trigger)
        self.observation_history = []

        self.aggregator = aggregator or _average_2d

    def compute_summary(self, trainer):
        if self.original_key in trainer.observation:
            value = trainer.observation[self.original_key]
            if isinstance(value, Variable):
                value.to_cpu()
            self.observation_history.append(value)

        if not self.comm_trigger(trainer):
            return None

        observation_history_gathered = self.comm.gather_obj(
            self.observation_history)
        self.observation_history = []

        if self.comm.rank == 0:
            global_summary = self.aggregator(observation_history_gathered)
            self.comm.bcast_obj(global_summary)
        else:
            global_summary = self.comm.bcast_obj(None)

        return global_summary

    def __call__(self, trainer):
        summary = self.compute_summary(trainer)

        if summary is not None:
            trainer.observation[self.aggregated_key] = summary


def _average_2d(xs):
    xs = sum(xs, [])
    return sum(xs) / len(xs)



from chainer.training.triggers import EarlyStoppingTrigger
from chainermn.extensions import ObservationAggregator


class MultiNodeEarlyStoppingTrigger(object):
    """__init__(\
        self, comm, check_trigger=(1, 'epoch'), monitor='main/loss', \
        patience=3, mode='auto', verbose=False, \
        max_trigger=(100, 'epoch'))

    Trigger for Early Stopping in Multiple Node Environments

    It serves almost the same as
    :class:`~chainer.training.triggers.EarlyStoppingTrigger`,
    but it can correctly work in multiple node environments.

    The difference between it and
    :class:`~chainer.training.triggers.EarlyStoppingTrigger` is that,
    in each check interval, it computes the mean of the accumulated
    values *across all nodes*. In this way, all nodes will have the same
    value to determine the timing at which the trigger fires so that
    they will stop at the same time.

    Args:
        comm : ChainerMN communicator
        check_trigger: Trigger that decides the comparison
            interval between current best value and new value.
            This must be a tuple in the form of ``<int>,
            'epoch'`` or ``<int>, 'iteration'`` which is passed to
            :class:`~chainer.training.triggers.IntervalTrigger`.
        monitor (str) : The metric you want to monitor
        patience (int) : Counts to let the trigger be patient.
            The trigger will not fire until the condition is met
            for successive ``patience`` checks.
        mode (str) : ``'max'``, ``'min'``, or ``'auto'``.
            It is used to determine how to compare the monitored values.
        verbose (bool) : Enable verbose output.
            If verbose is true, you can get more information
        max_trigger: Upper bound of the number of training loops
        suffix (str): Suffix added to the name of the monitored
            metric after aggregation.

    .. note::
       ``patients`` is also available as an alias of ``patience`` for
       historical reason.
    """

    def __init__(self, comm,
                 *, check_trigger=(1, 'epoch'), monitor='main/loss',
                 patience=None, mode='auto', verbose=False,
                 max_trigger=(100, 'epoch'), suffix='_aggregated', **kwargs):

        # `patients` as an alias of `patience`
        monitor_aggregated = monitor + suffix

        self.actual_trigger = EarlyStoppingTrigger(check_trigger=check_trigger,
                                                   monitor=monitor_aggregated,
                                                   patience=patience,
                                                   mode=mode, verbose=verbose,
                                                   max_trigger=max_trigger,
                                                   **kwargs)
        self.aggregator = ObservationAggregator(
            comm, monitor,
            aggregated_key=monitor_aggregated,
            comm_trigger=check_trigger)

    def __call__(self, trainer):
        self.aggregator(trainer)
        return self.actual_trigger(trainer)

    def _stop_condition(self):
        return self.actual_trigger._stop_condition()

    def _init_summary(self):
        return self.actual_trigger._init_summary()

    def get_training_length(self):
        return self.actual_trigger.get_training_length()



import copy
import six

from chainer.training import extension
from chainer import backend
from chainer.dataset import convert
from chainer import function
from chainer.utils import argument
import chainerx as chx


class GenericMultiNodeEvaluator(extension.Extension):
    '''Generic multi-node evaluator for non-allreducable evaluation.

    This is to evaluate a Dataset that cannot evenly divided across
    all processes in the communicator, for evaluation calculation that
    is not applicable to a simple add-and-devide style averaging among
    processes.

    Users are recommeneded to implement its own local calculation
    ``calc_local()`` (e.g.  at each distributed GPU) and aggregation
    ``aggregate()`` of its results. Although it has built-in
    implementaiton of those two methods.

    It has several drawbacks; 1) Additional implementation of
    aggregation required to users, and 2) no compatibility with
    :class:`~chainer.training.extensions.Evaluator`.

    .. note:: No automatic support of Reporter is provided; Set it up
       at ``initialize()`` method

    Args:
        comm:
            ChainerMN communicator object
        iterator:
            An iterator for test dataset. Must be non-repeated.
        target (callable):
            A model to evaluate with test dataset
        device (int or chainer.backend.Device):
            A device indicator to send data with converter. Not used
            when the converter is not using any devices.
        converter (callable):
            A converter. Default value is
            :func:`chainer.dataset.concat_examples` .
        root (int):
            Rank number of root process to run bcast and gather with.
        progress_hook (callable):
            A callable that receives single argument for indicators. The
            callable is only callled at root process.

    '''
    trigger = 1, 'epoch'
    default_name = 'validation'
    priority = extension.PRIORITY_WRITER

    name = None

    def __init__(self, comm, iterator, target, device=None,
                 converter=convert.concat_examples, root=0,
                 **kwargs):
        progress_hook, = argument.parse_kwargs(kwargs, ('progress_hook', None))

        self.comm = comm
        self.iterator = iterator
        self._targets = {"main": target}
        self.converter = converter

        if device is not None:
            device = backend.get_device(device)
        self.device = device

        self._progress_hook = progress_hook

        assert 0 <= root and root < self.comm.size
        self.root = root

    def __call__(self, trainer):
        if hasattr(self.iterator, 'reset'):
            self.iterator.reset()
            it = self.iterator
        else:
            it = copy.copy(self.iterator)

        if self.comm is not None:
            gen = self._evaluate_local(it)

            if self.comm.rank == self.root:
                total_result = self.aggregate([result for result in gen])
            else:
                for _ in gen:
                    pass
                total_result = None

        else:
            # Non-multinode environment
            gen = self._evaluate_local_single(self, it)
            total_result = self.aggregate([result for result in gen])

        return total_result

    def calc_local(self, *args, **kwargs):
        '''A generic method for local calculation.

        Override this method to run its local calculation.  Otherwise,
        results are calculated with original target and test dataset.

        Args:
            args:
                Result of converter when it is tuple.
            kwargs:
                Result of converter when it is dict.

        Returns:
            Arbrary value may be returned, but must not be ``None``.

        '''
        target = self._targets['main']
        return target(*args, **kwargs)

    def aggregate(self, results):
        '''A generic aggregation method.

        Override this method for original aggregation calculation. By
        default, it just does nothing but returns the input. This
        method is called once and only once across the cluster, at
        root process. Reporting can be run here.

        Args:
            results (list):
                List of return value of ``calc_local()`` obtained from
                all nodes..

        '''
        return results

    def _evaluate_local_single(self, iterator):
        for batch in iterator:
            in_arrays = convert._call_converter(
                self.converter, batch, self.device)

            with function.no_backprop_mode():
                if isinstance(in_arrays, tuple):
                    results = self.calc_local(*in_arrays)
                elif isinstance(in_arrays, dict):
                    results = self.calc_local(**in_arrays)
                else:
                    results = self.calc_local(in_arrays)

            if self._progress_hook:
                self._progress_hook(batch)
            yield results

    def _evaluate_local(self, iterator):
        # Check whether local eval is all done every 8 rounds
        gather_interval = 8

        all_done = None
        while not all_done:
            all_done = None
            results = None
            for _ in range(gather_interval):
                try:
                    batch = iterator.next()
                    in_arrays = convert._call_converter(
                        self.converter, batch, self.device)

                    with function.no_backprop_mode():
                        if isinstance(in_arrays, tuple):
                            results = self.calc_local(*in_arrays)
                        elif isinstance(in_arrays, dict):
                            results = self.calc_local(**in_arrays)
                        else:
                            results = self.calc_local(in_arrays)

                    if self.comm.rank == self.root and self._progress_hook:
                        self._progress_hook(batch)

                except StopIteration:
                    batch = None
                    results = None

                results = self.comm.gather_obj(results, root=self.root)

                if self.comm.rank == self.root:
                    valid_results = [r for r in results if r is not None]
                    for result in valid_results:
                        yield result

                    all_done = len(valid_results) == 0

            all_done = self.comm.bcast_obj(all_done, root=self.root)
        return


def create_multi_node_evaluator(actual_evaluator, communicator):
    """Create a multi node evaluator from a normal evaluator.

    Actually this method patches the evaluator to work in multi node
    environment. This method adds several hidden attributes starting
    with `_mn_` prefix.

    Args:
        actual_evaluator: evaluator to be patched
            (e.g., ``chainer.training.extensions.Evaluator``)
        communicator: ChainerMN communicator

    Returns:
        The multi-node patched ``actual_evaluator``.

    .. note:: After patched, original evaluator does not work
              correctly in non-MPI environment.

    """

    actual_evaluator._mn_original_evaluate = actual_evaluator.evaluate
    actual_evaluator._mn_communicator = communicator

    def new_evaluate(self):
        local_mean_dict = self._mn_original_evaluate()

        # ChainerX support:
        # We need convert chainerx ndarray to Native array because
        #   (1) allreduce_obj is used to compute global mean values, since
        #       a simple allreduce operation cannot be applied in evaluation.
        #   (2) allreduce_obj calls mpi4py.allreduce, which pickles the object
        #   (3) chainerx.ndarray preserves CUDA device internally when pickled
        #   (4) An error will occur when an ndarray is unpickled in another
        #       process
        array0 = list(local_mean_dict.values())[0]
        xp = backend.get_array_module(array0)
        if xp == chx and array0.device.backend.name == 'cuda':
            # Results of evaluation is fairly small, so
            # the ndarray is transferred to CPU and allreduce()-ed.
            # NOTE: Matrices for evaluation are transferred to the host memory
            # and sent via MPI instead of NCCL. Although evaluation matrices
            # are small in most cases, this is a potential performance issue.
            local_mean_dict = {
                name: chx.to_numpy(value)
                for name, value in local_mean_dict.items()
            }

        global_mean_dict = {
            name:
            self._mn_communicator.allreduce_obj(
                value) / self._mn_communicator.size
            for name, value in sorted(local_mean_dict.items())
        }
        return global_mean_dict

    actual_evaluator.evaluate = six.create_bound_method(
        new_evaluate, actual_evaluator)
    return actual_evaluator



import numpy

from chainer.backends import cuda
from chainer import initializer

# Original code from Berkeley FCN
# https://github.com/shelhamer/fcn.berkeleyvision.org/blob/master/surgery.py


def _get_linear_filter(size, ndim, upsampling=True):
    """Make a 2D and 3D linear kernel suitable for up/downsampling"""
    factor = (size + 1) // 2
    if size % 2 == 1:
        center = factor - 1.
    else:
        center = factor - 0.5
    slices = (slice(size),) * ndim
    og = numpy.ogrid[slices]
    filt = 1.
    for og_i in og:
        filt = filt * (1. - abs(og_i - center) / factor)
    if not upsampling:
        filt /= filt.sum()
    return filt


class _SamplingFilter(initializer.Initializer):

    def __init__(self, upsampling=True, interpolation='linear', dtype=None):
        self._upsampling = upsampling
        if interpolation == 'linear':
            self._get_filter_func = _get_linear_filter
        else:
            raise ValueError(
                'Unsupported interpolation method: {}'.format(interpolation))
        super(_SamplingFilter, self).__init__(dtype)

    def __call__(self, array):
        if self.dtype is not None:
            assert array.dtype == self.dtype
        xp = cuda.get_array_module(array)

        in_c, out_c = array.shape[:2]
        assert in_c == out_c or out_c == 1

        ksize = None
        for k in array.shape[2:]:
            if ksize is None:
                ksize = k
            else:
                if ksize != k:
                    raise ValueError(
                        'ksize must be all same: {} != {}'.format(ksize, k))

        filt = self._get_filter_func(
            ksize, ndim=array.ndim - 2, upsampling=self._upsampling)
        filt = xp.asarray(filt)

        array[...] = 0.
        if out_c == 1:
            array[xp.arange(in_c), 0, ...] = filt
        else:
            array[xp.arange(in_c), xp.arange(out_c), ...] = filt


class UpsamplingDeconvFilter(_SamplingFilter):

    """Initializes array with upsampling filter.

    The array is initialized with a standard image upsampling weight.
    This initializer is often used as initial weight for
    :func:`~chainer.links.DeconvolutionND`.
    :func:`~chainer.links.DeconvolutionND` is expected that its `stride` is
    equal to `(ksize + 1) // 2`.

    Reference: Long et al., https://arxiv.org/abs/1411.4038

    Attributes:
        interpolation (str): Upsampling interpolation method.
        Default is 'linear'.

    """

    def __init__(self, interpolation='linear', dtype=None):
        if interpolation != 'linear':
            raise ValueError(
                'Unsupported interpolation method: {}'.format(interpolation))
        super(UpsamplingDeconvFilter, self).__init__(
            upsampling=True, interpolation=interpolation, dtype=dtype)


class DownsamplingConvFilter(_SamplingFilter):

    """Initializes array with downsampling filter.

    The array is initialized with a standard image downsampling weight.
    This initializer is often used as initial weight for
    :func:`~chainer.links.ConvolutionND`.
    :func:`~chainer.links.ConvolutionND` is expected that its `stride` is
    equal to `(ksize + 1) // 2`.

    Reference: Long et al., https://arxiv.org/abs/1411.4038

    Attributes:
        interpolation (str): Downsampling interpolation method.
        Default is 'linear'.

    """

    def __init__(self, interpolation='linear', dtype=None):
        if interpolation != 'linear':
            raise ValueError(
                'Unsupported interpolation method: {}'.format(interpolation))
        super(DownsamplingConvFilter, self).__init__(
            upsampling=False, interpolation=interpolation, dtype=dtype)



from onnx_chainer.functions.opset_version import support
from onnx_chainer import onnx_helper


def _convert_softmax_impl(op_type, func, input_names, output_names):
    axis = func.axis
    ndim = len(func.inputs[0].shape)
    if axis == ndim - 1:
        return onnx_helper.make_node(
            op_type, input_names, output_names,
            axis=axis
        ),

    # Chainer's softmax computes the softmax along a single axis while
    # ONNX's computes along the specified axis and all axes after the
    # specified axis. To emulate Chainer's by ONNX's, we transpose the
    # single specified axis to the last axis, compute the softmax, and
    # transpose back to the original shape.
    gb = onnx_helper.GraphBuilder()
    perm = list(range(ndim))
    perm[axis], perm[-1] = perm[-1], perm[axis]
    transposed = gb.op('Transpose', input_names, perm=perm)
    softmaxed = gb.op(op_type, [transposed], axis=ndim - 1)
    gb.op('Transpose', [softmaxed], perm=perm)
    return gb.nodes(output_names=output_names)


@support((1, 6))
def convert_ClippedReLU(
        func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Clip', input_names, output_names,
            min=0.0, max=func.cap,
            consumed_inputs=[1]
        ),
    elif opset_version == 6:
        return onnx_helper.make_node(
            'Clip', input_names, output_names,
            min=0.0, max=func.cap,
        ),


@support((1, 6))
def convert_ELU(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Elu', input_names, output_names,
            alpha=func.alpha,
        ),
    elif opset_version == 6:
        return onnx_helper.make_node(
            'Elu', input_names, output_names,
            alpha=func.alpha
        ),


@support((1, 6))
def convert_HardSigmoid(
        func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'HardSigmoid', input_names, output_names,
            alpha=0.2,
            beta=0.5,
            consumed_inputs=[1],
        ),
    elif opset_version == 6:
        return onnx_helper.make_node(
            'HardSigmoid', input_names, output_names,
            alpha=0.2,
            beta=0.5
        ),


@support((1, 6))
def convert_LeakyReLU(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'LeakyRelu', input_names, output_names,
            alpha=func.slope,
            consumed_inputs=[1],
        ),
    elif opset_version == 6:
        return onnx_helper.make_node(
            'LeakyRelu', input_names, output_names,
            alpha=func.slope
        ),


def convert_LogSoftmax(
        func, opset_version, input_names, output_names, context):
    return _convert_softmax_impl('LogSoftmax', func, input_names, output_names)


@support((1, 6, 7))
def convert_PReLUFunction(
        func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'PRelu', input_names, output_names, consumed_inputs=[1]),
    elif opset_version == 6:
        return onnx_helper.make_node('PRelu', input_names, output_names),
    elif opset_version == 7:
        return onnx_helper.make_node('PRelu', input_names, output_names),


@support((1, 6))
def convert_ReLU(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Relu', input_names, output_names,
            consumed_inputs=[1]),
    elif opset_version == 6:
        return onnx_helper.make_node('Relu', input_names, output_names),


@support((1, 6))
def convert_Selu(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Selu', input_names, output_names,
            consumed_inputs=[1],
            alpha=func.alpha,
            gamma=func.scale
        ),
    elif opset_version == 6:
        return onnx_helper.make_node('Selu', input_names, output_names,
                                     alpha=func.alpha,
                                     gamma=func.scale
                                     ),


@support((1, 6))
def convert_Sigmoid(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Sigmoid', input_names, output_names,
            consumed_inputs=[1]),
    elif opset_version == 6:
        return onnx_helper.make_node('Sigmoid', input_names, output_names),


def convert_Softmax(func, opset_version, input_names, output_names, context):
    return _convert_softmax_impl('Softmax', func, input_names, output_names)


def convert_Softplus(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node('Softplus', input_names, output_names),


@support((1, 6))
def convert_Tanh(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Tanh', input_names, output_names,
            consumed_inputs=[1]),
    elif opset_version == 6:
        return onnx_helper.make_node('Tanh', input_names, output_names),



import numpy

from chainer.backend import cuda
from chainer import distribution
from chainer.functions.array import repeat
from chainer.functions.array import reshape
from chainer.functions.array import transpose
from chainer.functions.math import prod
from chainer.functions.math import sum as sum_mod
from chainer.utils import array
from chainer.utils import cache


class Independent(distribution.Distribution):

    """Independent distribution.

    Args:
        distribution (:class:`~chainer.Distribution`): The base distribution
            instance to transform.
        reinterpreted_batch_ndims (:class:`int`): Integer number of rightmost
            batch dims which will be regarded as event dims. When ``None`` all
            but the first batch axis (batch axis 0) will be transferred to
            event dimensions.
    """

    def __init__(self, distribution, reinterpreted_batch_ndims=None):
        super(Independent, self).__init__()
        self.__distribution = distribution
        if reinterpreted_batch_ndims is None:
            reinterpreted_batch_ndims = (
                self._get_default_reinterpreted_batch_ndims(distribution))
        elif reinterpreted_batch_ndims > len(distribution.batch_shape):
            raise ValueError(
                'reinterpreted_batch_ndims must be less than or equal to the '
                'number of dimensions of `distribution.batch_shape`.')
        self.__reinterpreted_batch_ndims = reinterpreted_batch_ndims

        batch_ndim = (
            len(self.distribution.batch_shape)
            - self.reinterpreted_batch_ndims)
        self.__batch_shape = distribution.batch_shape[:batch_ndim]
        self.__event_shape = (
            distribution.batch_shape[batch_ndim:]
            + distribution.event_shape)

    @property
    def distribution(self):
        return self.__distribution

    @property
    def reinterpreted_batch_ndims(self):
        return self.__reinterpreted_batch_ndims

    @property
    def batch_shape(self):
        return self.__batch_shape

    @property
    def event_shape(self):
        return self.__event_shape

    @cache.cached_property
    def covariance(self):
        """ The covariance of the independent distribution.

        By definition, the covariance of the new
        distribution becomes block diagonal matrix. Let
        :math:`\\Sigma_{\\mathbf{x}}` be the covariance matrix of the original
        random variable :math:`\\mathbf{x} \\in \\mathbb{R}^d`, and
        :math:`\\mathbf{x}^{(1)}, \\mathbf{x}^{(2)}, \\cdots \\mathbf{x}^{(m)}`
        be the :math:`m` i.i.d. random variables, new covariance matrix
        :math:`\\Sigma_{\\mathbf{y}}` of :math:`\\mathbf{y} =
        [\\mathbf{x}^{(1)}, \\mathbf{x}^{(2)}, \\cdots, \\mathbf{x}^{(m)}] \\in
        \\mathbb{R}^{md}` can be written as

        .. math::
            \\left[\\begin{array}{ccc}
                    \\Sigma_{\\mathbf{x}^{1}} & & 0 \\\\
                    & \\ddots & \\\\
                    0 & & \\Sigma_{\\mathbf{x}^{m}}
            \\end{array} \\right].

        Note that this relationship holds only if the covariance matrix of the
        original distribution is given analytically.

        Returns:
            ~chainer.Variable: The covariance of the distribution.
        """
        n_repeats = array.size_of_shape(
            self.distribution.batch_shape[-self.reinterpreted_batch_ndims:])
        dim = array.size_of_shape(self.distribution.event_shape)
        cov = repeat.repeat(
            reshape.reshape(
                self.distribution.covariance,
                ((self.batch_shape) + (1, n_repeats, dim, dim))),
            n_repeats, axis=-4)
        cov = reshape.reshape(
            transpose.transpose(
                cov, axes=(
                    tuple(range(len(self.batch_shape))) + (-4, -2, -3, -1))),
            self.batch_shape + (n_repeats * dim, n_repeats * dim))
        block_indicator = self.xp.reshape(
            self._block_indicator,
            tuple([1] * len(self.batch_shape)) + self._block_indicator.shape)
        return cov * block_indicator

    @property
    def entropy(self):
        return self._reduce(sum_mod.sum, self.distribution.entropy)

    def cdf(self, x):
        return self._reduce(prod.prod, self.distribution.cdf(x))

    def icdf(self, x):
        """The inverse cumulative distribution function for multivariate variable.

        Cumulative distribution function for multivariate variable is not
        invertible. This function always raises :class:`RuntimeError`.

        Args:
            x (:class:`~chainer.Variable` or :ref:`ndarray`): Data points in
                the codomain of the distribution

        Raises:
            :class:`RuntimeError`
        """

        raise RuntimeError(
            'Cumulative distribution function for multivariate variable '
            'is not invertible.')

    def log_cdf(self, x):
        return self._reduce(sum_mod.sum, self.distribution.log_cdf(x))

    def log_prob(self, x):
        return self._reduce(sum_mod.sum, self.distribution.log_prob(x))

    def log_survival_function(self, x):
        return self._reduce(
            sum_mod.sum, self.distribution.log_survival_function(x))

    @property
    def mean(self):
        return self.distribution.mean

    @property
    def mode(self):
        return self.distribution.mode

    @property
    def params(self):
        return self.distribution.params

    def perplexity(self, x):
        return self._reduce(prod.prod, self.distribution.perplexity(x))

    def prob(self, x):
        return self._reduce(prod.prod, self.distribution.prob(x))

    def sample_n(self, n):
        return self.distribution.sample_n(n)

    @property
    def stddev(self):
        return self.distribution.stddev

    @property
    def support(self):
        return self.distribution.support

    def survival_function(self, x):
        return self._reduce(prod.prod, self.distribution.survival_function(x))

    @property
    def variance(self):
        return self.distribution.variance

    @property
    def xp(self):
        return self.distribution.xp

    def _reduce(self, op, stat):
        range_ = tuple(range(-self.reinterpreted_batch_ndims, 0))
        return op(stat, axis=range_)

    def _get_default_reinterpreted_batch_ndims(self, distribution):
        ndims = len(distribution.batch_shape)
        return max(0, ndims - 1)

    @cache.cached_property
    def _block_indicator(self):
        n_repeats = array.size_of_shape(
            self.distribution.batch_shape[-self.reinterpreted_batch_ndims:])
        dim = array.size_of_shape(self.distribution.event_shape)
        block_indicator = numpy.fromfunction(
            lambda i, j: i // dim == j // dim,
            (n_repeats * dim, n_repeats * dim)).astype(int)
        if self.xp is cuda.cupy:
            block_indicator = cuda.to_gpu(block_indicator)
        return block_indicator


@distribution.register_kl(Independent, Independent)
def _kl_independent_independent(dist1, dist2):
    """Computes Kullback-Leibler divergence for independent distributions.

    We can leverage the fact that
    .. math::
        \\mathrm{KL}(
                \\mathrm{Independent}(\\mathrm{dist1}) ||
                \\mathrm{Independent}(\\mathrm{dist2}))
        = \\mathrm{sum}(\\mathrm{KL}(\\mathrm{dist1} || \\mathrm{dist2}))
    where the sum is over the ``reinterpreted_batch_ndims``.

    Args:
        dist1 (:class:`~chainer.distribution.Independent`): Instance of
            `Independent`.
        dist2 (:class:`~chainer.distribution.Independent`): Instance of
            `Independent`.

    Returns:
        Batchwise ``KL(dist1 || dist2)``.

    Raises:
        :class:`ValueError`: If the event space for ``dist1`` and ``dist2``,
            or their underlying distributions don't match.
    """

    p = dist1.distribution
    q = dist2.distribution

    # The KL between any two (non)-batched distributions is a scalar.
    # Given that the KL between two factored distributions is the sum, i.e.
    # KL(p1(x)p2(y) || q1(x)q2(y)) = KL(p1 || q1) + KL(q1 || q2), we compute
    # KL(p || q) and do a `reduce_sum` on the reinterpreted batch dimensions.
    if dist1.event_shape == dist2.event_shape:
        if p.event_shape == q.event_shape:
            n_axes = len(dist1.event_shape) - len(p.event_shape)
            n_reduce_dims = tuple([-i - 1 for i in range(0, n_axes)])

            return sum_mod.sum(
                distribution.kl_divergence(p, q), axis=n_reduce_dims)
        else:
            raise NotImplementedError(
                'KL between Independents with different '
                'event shapes not supported.')
    else:
        raise ValueError('Event shapes do not match.')



import collections
import contextlib
import shutil
import sys
import tempfile

import numpy
import six

import chainer
# import classes and functions
from chainer.utils.array import size_of_shape  # NOQA
from chainer.utils.array import sum_to  # NOQA
from chainer.utils.conv import get_conv_outsize  # NOQA
from chainer.utils.conv import get_deconv_outsize  # NOQA
from chainer.utils.error import _format_array_props  # NOQA
from chainer.utils.experimental import experimental  # NOQA
from chainer.utils.meta import enable_final  # NOQA
from chainer.utils.meta import final  # NOQA
from chainer.utils.nondeterministic import nondeterministic  # NOQA
from chainer.utils.sparse import CooMatrix  # NOQA
from chainer.utils.sparse import get_order  # NOQA
from chainer.utils.sparse import to_coo  # NOQA

# The following alias has been moved to chainer/__init__.py in order to break
# circular imports in Python 2.
# from chainer.utils.walker_alias import WalkerAlias


# TODO(kmaehashi) remove this when `six.moves.collections_abc` is implemented.
# See: https://github.com/chainer/chainer/issues/5097
try:
    collections_abc = collections.abc  # type: ignore
except AttributeError:  # python <3.3
    collections_abc = collections  # type: ignore


def force_array(x, dtype=None):
    # numpy returns a float value (scalar) when a return value of an operator
    # is a 0-dimension array.
    # We need to convert such a value to a 0-dimension array because `Function`
    # object needs to return an `numpy.ndarray`.
    if numpy.isscalar(x):
        if dtype is None:
            return numpy.array(x)
        else:
            return numpy.array(x, dtype)
    else:
        if dtype is None:
            return x
        else:
            return x.astype(dtype, copy=False)


def force_type(dtype, value):
    if numpy.isscalar(value):
        return dtype.type(value)
    elif value.dtype != dtype:
        return value.astype(dtype, copy=False)
    else:
        return value


@contextlib.contextmanager
def tempdir(**kwargs):
    # A context manager that defines a lifetime of a temporary directory.
    ignore_errors = kwargs.pop('ignore_errors', False)

    temp_dir = tempfile.mkdtemp(**kwargs)
    try:
        yield temp_dir
    finally:
        shutil.rmtree(temp_dir, ignore_errors=ignore_errors)


def _repr_with_named_data(inst, **kwargs):
    """Convenient function to generate `repr` string with custom named data"""
    if six.PY2:
        class_name = inst.__class__.__name__
    else:
        class_name = inst.__class__.__qualname__
    return '<{}.{} {}>'.format(
        inst.__module__, class_name,
        ' '.join('{}={}'.format(k, v) for k, v in six.iteritems(kwargs)))


def _check_arrays_forward_compatible(arrays, label=None):
    if not chainer.is_arrays_compatible(arrays):
        raise TypeError(
            'incompatible array types are mixed in the forward input{}.\n'
            'Actual: {}'.format(
                ' ({})'.format(label) if label is not None else '',
                ', '.join(str(type(a)) for a in arrays)))


def _raise_from(exc_type, message, orig_exc):
    # Raises an exception that wraps another exception.
    message = (
        '{}\n\n'
        '(caused by)\n'
        '{}: {}\n'.format(message, type(orig_exc).__name__, orig_exc))
    new_exc = exc_type(message)
    if sys.version_info < (3,):
        six.reraise(exc_type, new_exc, sys.exc_info()[2])
    else:
        six.raise_from(new_exc.with_traceback(orig_exc.__traceback__), None)



import warnings

import chainer
import numpy as np
from onnx.mapping import NP_TYPE_TO_TENSOR_TYPE

from onnx_chainer.functions.opset_version import support
from onnx_chainer import onnx_helper


TENSOR_TYPE_TO_NAME = {
    0: 'UNDEFINED',
    1: 'FLOAT',
    2: 'UINT8',
    3: 'INT8',
    4: 'UINT16',
    5: 'INT16',
    6: 'INT32',
    7: 'INT64',
    8: 'STRING',
    9: 'BOOL',
    10: 'FLOAT16',
    11: 'DOUBLE',
    12: 'UINT32',
    13: 'UINT64',
    14: 'COMPLEX64',
    15: 'COMPLEX128',
}


@support((1, 6))
def convert_Cast(func, opset_version, input_names, output_names, context):
    typ = func.type if isinstance(func.type, np.dtype) else np.dtype(func.type)
    if opset_version == 1:
        return onnx_helper.make_node(
            'Cast', input_names, output_names,
            to=TENSOR_TYPE_TO_NAME[NP_TYPE_TO_TENSOR_TYPE[typ]]
        ),
    elif opset_version == 6:
        return onnx_helper.make_node(
            'Cast', input_names, output_names,
            to=NP_TYPE_TO_TENSOR_TYPE[typ]
        ),


@support((1, 4))
def convert_Concat(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Concat', input_names, output_names,
            axis=func.axis
        ),
    elif opset_version == 4:
        return onnx_helper.make_node(
            'Concat', input_names, output_names,
            axis=func.axis
        ),


def convert_Copy(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node(
        'Identity', input_names, output_names
    ),


def convert_Depth2Space(
        func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node(
        'DepthToSpace', input_names, output_names,
        blocksize=func.r
    ),


def get_slice_node(
        gb, opset_version, context, input_names, axes, starts, ends):
    if opset_version < 10:
        return gb.op(
            'Slice', input_names, axes=axes, starts=starts, ends=ends)
    else:
        for param in [('starts', starts), ('ends', ends), ('axes', axes)]:
            param_name = context.add_const(
                np.asarray(list(param[1]), dtype=np.int64), param[0])
            input_names.append(param_name)
        return gb.op('Slice', input_names)


@support((1, 10))
def convert_GetItem(func, opset_version, input_names, output_names, context):
    x = func.inputs[0]
    axes, starts, ends = [], [], []
    squeeze_idxs, unsqueeze_idxs = [], []
    skipped = 0  # when set ellipsis, need to skip index rolling

    gather_axis, gather_idx = [], []

    for i, idx in enumerate(func.slices):
        # axis means the index of input x, adjust None and Ellipsis counts
        axis = i - len(unsqueeze_idxs) + skipped
        if isinstance(idx, slice):
            if idx.step is not None and idx.step != 1:
                raise ValueError(
                    'GetItem with {}step slicing is not supported in ONNX '
                    'Slice operator'.format(idx.step))
            if idx.start is None and idx.stop is None:
                continue
            axes.append(axis)
            starts.append(0 if idx.start is None else idx.start)
            ends.append(x.shape[axis] if idx.stop is None else idx.stop)
        elif isinstance(idx, int):
            axes.append(axis)
            starts.append(idx)
            ends.append(idx+1)
            squeeze_idxs.append(axis)
        elif isinstance(idx, np.ndarray) and idx.ndim == 0:
            scalar_idx = idx.item()
            axes.append(axis)
            starts.append(scalar_idx)
            ends.append(scalar_idx+1)
            squeeze_idxs.append(axis)
        elif idx is None:
            unsqueeze_idxs.append(i - len(squeeze_idxs) + skipped)
        elif idx is Ellipsis:
            # calculate rest slice number except None, GetItem does not allow
            # multiple Ellipsis, so ignore latter Ellipsis count
            rest_slice_len = len(
                [idx_ for idx_ in func.slices[i+1:] if idx_ is not None])
            assert skipped == 0
            skipped = len(x.shape) - axis - rest_slice_len - 1
        elif isinstance(idx, (list,) + chainer.get_array_types()):
            if gather_axis:
                raise ValueError(
                    'ONNX-Chainer does not support multiple advanced index')
            gather_axis.append(axis - len(squeeze_idxs) + len(unsqueeze_idxs))
            if isinstance(idx, list):
                gather_idx = np.array(idx, dtype=np.int64)
            else:
                gather_idx = chainer.cuda.to_cpu(idx)
        else:
            # not support advanced index like `array[[0,1], [0, 1]]`
            raise ValueError(
                'GetItem with type {} cannot handle in ONNX Slice, so that '
                'ONNX-Chainer does not accept the type'.format(type(idx)))

    gb = onnx_helper.GraphBuilder()
    slice_output = input_names
    if axes:
        output = get_slice_node(
            gb, opset_version, context, slice_output, axes, starts, ends)
        slice_output = [output]
    if squeeze_idxs:
        output = gb.op('Squeeze', slice_output, axes=squeeze_idxs)
        slice_output = [output]
    if unsqueeze_idxs:
        output = gb.op('Unsqueeze', slice_output, axes=unsqueeze_idxs)
        slice_output = [output]

    if gather_axis:
        gather_idx_name = context.add_const(gather_idx, 'indices')
        slice_output.append(gather_idx_name)
        gb.op('Gather', slice_output, axis=gather_axis[0])

    return gb.nodes(output_names=output_names)


@support((1, 2))
def convert_Pad(func, opset_version, input_names, output_names, context):
    if func.mode not in ['constant', 'reflect', 'edge']:
        raise ValueError(
            '{} mode is not supported in ONNX\'s Pad operation'.format(
                func.mode))

    pad_begin = []
    pad_end = []
    pad_bw = func.pad_bw
    if pad_bw.ndim == 1:
        pad_bw = np.tile(pad_bw, (len(func.inputs[0].shape), 1))
    for pp in pad_bw.tolist():
        pad_begin.append(pp[0])
        pad_end.append(pp[1])
    pad = pad_begin + pad_end

    if 'constant_values' in func.keywords:
        values = func.keywords['constant_values']
        if not isinstance(values, int) and len(values) > 1:
            raise ValueError(
                'ONNX doesn\'t support multiple constant values for Pad '
                'operation')
        elif not isinstance(values, int):
            values = float(values[0])
        else:
            values = float(values)

        if opset_version == 1:
            node = onnx_helper.make_node(
                'Pad', input_names, output_names,
                mode=func.mode,
                paddings=pad,
                value=values
            )
        elif opset_version == 2:
            node = onnx_helper.make_node(
                'Pad', input_names, output_names,
                mode=func.mode,
                pads=pad,
                value=values
            )
    else:
        if opset_version == 1:
            node = onnx_helper.make_node(
                'Pad', input_names, output_names,
                mode=func.mode,
                paddings=pad,
                value=0.,
            )
        elif opset_version == 2:
            node = onnx_helper.make_node(
                'Pad', input_names, output_names,
                mode=func.mode,
                pads=pad,
                value=0.,
            )

    return node,


@support((1, 5))
def convert_Reshape(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Reshape', input_names, output_names,
            shape=func.shape
        ),
    elif opset_version == 5:
        if hasattr(func, 'shape'):
            # if the function has shape parameter, means not dynamic
            assert len(input_names) == 1
            shape_name = context.add_const(
                np.asarray(list(func.shape), dtype=np.int64), 'shape')
            input_names.append(shape_name)
        else:
            if len(input_names) != 2:
                raise ValueError('shape must be set as parameter or 2nd input')

        return onnx_helper.make_node(
            'Reshape', input_names, output_names,
        ),


def convert_Space2Depth(
        func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node(
        'SpaceToDepth', input_names, output_names,
        blocksize=func.r
    ),


@support((1, 2))
def convert_SplitAxis(func, opset_version, input_names, output_names, context):
    if func.indices is not None:
        indices_or_sections = func.indices
    else:
        indices_or_sections = func.sections

    total = func.inputs[0].shape[func.axis]
    if hasattr(indices_or_sections, '__iter__'):
        split = []
        prev_i = 0
        for i in indices_or_sections:
            split.append(i - prev_i)
            prev_i = i
        split.append(total - prev_i)
    else:
        length = total // indices_or_sections
        split = [length for _ in range(indices_or_sections)]

    assert len(output_names) == len(split)
    if opset_version == 1:
        return onnx_helper.make_node(
            'Split', input_names, output_names,
            axis=func.axis,
            split=split
        ),
    elif opset_version == 2:
        return onnx_helper.make_node(
            'Split', input_names, output_names,
            axis=func.axis,
            split=split
        ),


def convert_Squeeze(func, opset_version, input_names, output_names, context):
    if func.axis is None:
        axis = []
        for i, s in enumerate(func.inputs[0].shape):
            if s == 1:
                axis.append(i)
    else:
        axis = func.axis

    return onnx_helper.make_node(
        'Squeeze', input_names, output_names,
        axes=axis
    ),


def convert_Swapaxes(func, opset_version, input_names, output_names, context):
    perm = list(range(len(func.inputs[0].shape)))
    perm[func.axis1], perm[func.axis2] = perm[func.axis2], perm[func.axis1]

    return onnx_helper.make_node(
        'Transpose', input_names, output_names, perm=perm
    ),


@support((1, 6))
def convert_Tile(func, opset_version, input_names, output_names, context):
    # Add tiles and axis to graph
    if isinstance(func.reps, int):
        func.reps = [func.reps]
    tiles_name = context.add_const(
        np.asarray(func.reps, dtype=np.int64), 'tiles')
    input_names.append(tiles_name)

    # In operater version = 1, axis also should be given
    if opset_version == 1:
        axis_name = context.add_const(
            np.array([i for i, _ in enumerate(func.reps)], dtype=np.float32),
            'axis')
        input_names.append(axis_name)

    return onnx_helper.make_node('Tile', input_names, output_names),


def convert_Transpose(func, opset_version, input_names, output_names, context):

    if func.axes is None:
        node = onnx_helper.make_node('Transpose', input_names, output_names)
    else:
        node = onnx_helper.make_node(
            'Transpose', input_names, output_names,
            perm=func.axes
        )

    return node,


def convert_ExpandDims(
        func, opset_version, input_names, output_names, context):
    axis = func.axis
    if axis < 0:
        axis = len(func.inputs[0].shape) + 1 + axis

    return onnx_helper.make_node(
        'Unsqueeze', input_names, output_names, axes=[axis]),


@support((9,))
def convert_Where(func, opset_version, input_names, output_names, context):
    input_names.insert(0, context.get_name(func.condition))
    return onnx_helper.make_node('Where', input_names, output_names),


@support((7, 9, 10))
def convert_Repeat(func, opset_version, input_names, output_names, context):
    repeats = func.repeats
    if len(repeats) > 1:
        raise NotImplementedError(
            'ONNX-Chainer currently does not support elementwise repeat')

    gb = onnx_helper.GraphBuilder()
    inputs = list(input_names)
    axis = func.axis
    if axis is None:
        shape_name = context.add_const(np.array([-1], dtype=np.int64), 'shape')
        input_names.append(shape_name)
        inputs = [gb.op('Reshape', input_names)]
        scales = [float(repeats[0])]
    else:
        scales = [1.0] * func.inputs[0].data.ndim
        scales[axis] = float(repeats[0])

    if opset_version == 7:
        gb.op_output_named('Upsample', inputs, output_names, scales=scales)
        return gb.nodes()

    if opset_version in [9, 10]:
        scales_name = context.add_const(
            np.array(scales, dtype=np.float32), 'scales')
        inputs.append(scales_name)
        op = 'Upsample' if opset_version == 9 else 'Resize'
        gb.op_output_named(op, inputs, output_names)
        return gb.nodes()


@support((7, 9, 10))
def convert_ResizeImages(
        func, opset_version, input_names, output_names, context):

    warnings.warn(
        '`resize_images` is mapped to `Upsampling` ONNX op with bilinear '
        'interpolation. '
        'Behavior of bilinear interpolation differs from each implementation. '
        'See the issue https://github.com/chainer/onnx-chainer/issues/147 '
        'for details.',
        UserWarning)

    outsize = (func.out_H, func.out_W)

    h, w = func.inputs[0].shape[2:]

    # Compute scaling factor.
    # NOTE(syoyo): Despite of its name, `Upsample` onnx op will downsample
    # images when scale value is less than 1.0
    scales = [1.0, 1.0, float(outsize[0]) / float(h),
              float(outsize[1]) / float(w)]

    if (scales[2] < 1.0e-8) and (scales[3] < 1.0e-8):
        raise ValueError(
            'scaling factor is too small or zero. scales for h = {}, scales '
            'for w = {}'.format(scales[2], scales[3]))

    # resize_images in Chainer only supports bilinear interpolation
    # Actually this will be mapped to 'bilinear' in onnxruntime
    mode = 'linear'
    if opset_version == 7:
        return onnx_helper.make_node('Upsample', input_names, output_names,
                                     scales=scales, mode=mode),

    if opset_version in [9, 10]:
        scales_name = context.add_const(
            np.array(scales, dtype=np.float32), 'scales')
        input_names.append(scales_name)
        op = 'Upsample' if opset_version == 9 else 'Resize'
        return onnx_helper.make_node(op, input_names, output_names,
                                     mode=mode),


def convert_Stack(func, opset_version, input_names, output_names, context):
    gb = onnx_helper.GraphBuilder()
    axis = func.axis
    if axis < 0:
        axis = len(func.inputs[0].shape) + 1 + axis

    # To use concat op, reshape every inputs add new axes
    inputs = [gb.op('Unsqueeze', [name], axes=[axis]) for name in input_names]
    gb.op_output_named('Concat', inputs, output_names, axis=axis)
    return gb.nodes()


def convert_Hstack(func, opset_version, input_names, output_names, context):
    gb = onnx_helper.GraphBuilder()
    input0_ndim = len(func.inputs[0].shape)
    inputs = input_names
    axis = 1
    if input0_ndim == 0:
        inputs = [gb.op('Unsqueeze', [name], axes=[0]) for name in input_names]
        axis = 0
    elif input0_ndim == 1:
        axis = 0
    gb.op_output_named('Concat', inputs, output_names, axis=axis)
    return gb.nodes()


def convert_Vstack(func, opset_version, input_names, output_names, context):
    gb = onnx_helper.GraphBuilder()
    input0_ndim = len(func.inputs[0].shape)
    inputs = input_names
    if input0_ndim == 0:
        inputs = [gb.op('Unsqueeze', [name], axes=[0, 1]) for
                  name in input_names]
    elif input0_ndim == 1:
        inputs = [gb.op('Unsqueeze', [name], axes=[0]) for name in input_names]
    gb.op_output_named('Concat', inputs, output_names, axis=0)
    return gb.nodes()


def convert_Dstack(func, opset_version, input_names, output_names, context):
    gb = onnx_helper.GraphBuilder()
    input0_ndim = len(func.inputs[0].shape)
    inputs = input_names
    if input0_ndim == 0:
        inputs = [gb.op('Unsqueeze', [name], axes=[0, 1, 2]) for
                  name in input_names]
    elif input0_ndim == 1:
        inputs = [gb.op('Unsqueeze', [name], axes=[0, 2]) for
                  name in input_names]
    elif input0_ndim == 2:
        inputs = [gb.op('Unsqueeze', [name], axes=[2]) for name in input_names]
    gb.op_output_named('Concat', inputs, output_names, axis=2)
    return gb.nodes()


def convert_Separate(func, opset_version, input_names, output_names, context):
    gb = onnx_helper.GraphBuilder()
    split_outs = gb.op(
        'Split', input_names, num_outputs=len(output_names), axis=func.axis)
    for i, node_name in enumerate(split_outs):
        gb.op_output_named(
            'Squeeze', [node_name], [output_names[i]], axes=[func.axis])
    return gb.nodes()


def convert_Shape(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node('Shape', input_names, output_names),


def convert_Moveaxis(func, opset_version, input_names, output_names, context):

    ndim = len(func.inputs[0].shape)
    source = [a % ndim for a in func.source]
    destination = [a % ndim for a in func.destination]

    order = [n for n in range(ndim) if n not in source]
    for dest, src in sorted(zip(destination, source)):
        order.insert(dest, src)

    node = onnx_helper.make_node('Transpose', input_names, output_names,
                                 perm=order)

    return node,



from onnx_chainer import functions
from onnx_chainer.functions.converter import FunctionConverter


_supported_function_node_set = {
    # Activation
    'ClippedReLU',
    'ELU',
    'HardSigmoid',
    'LeakyReLU',
    'LogSoftmax',
    'PReLUFunction',
    'ReLU',
    'Selu',
    'Sigmoid',
    'Softmax',
    'Softplus',
    'Tanh',

    # Array
    'Cast',
    'Concat',
    'Copy',
    'Depth2Space',
    'Dstack',
    'ExpandDims',
    'GetItem',
    'Hstack',
    'Moveaxis',
    'Pad',
    'Repeat',
    'Reshape',
    'ResizeImages',
    'Separate',
    'Shape',
    'Space2Depth',
    'SplitAxis',
    'Squeeze',
    'Stack',
    'Swapaxes',
    'Tile',
    'Transpose',
    'Vstack',
    'Where',

    # Connection
    'Convolution2DFunction',
    'ConvolutionND',
    'Deconvolution2DFunction',
    'DeconvolutionND',
    'EmbedIDFunction',
    'LinearFunction',

    # Loss
    'SoftmaxCrossEntropy',

    # Math
    'Absolute',
    'Add',
    'AddConstant',
    'Arccos',
    'Arcsin',
    'Arctan',
    'ArgMax',
    'ArgMin',
    'BroadcastTo',
    'Clip',
    'Cos',
    'Cosh',
    'Div',
    'DivFromConstant',
    'Exp',
    'Identity',
    'LinearInterpolate',
    'Log',
    'LogSumExp',
    'MatMul',
    'Max',
    'Maximum',
    'Mean',
    'Min',
    'Minimum',
    'Mul',
    'MulConstant',
    'Neg',
    'PowConstVar',
    'PowVarConst',
    'PowVarVar',
    'Prod',
    'RsqrtGPU',
    'Sin',
    'Sinh',
    'Sqrt',
    'Square',
    'Sub',
    'SubFromConstant',
    'Sum',
    'Tan',

    # Noise
    'Dropout',

    # Normalization
    'BatchNormalization',
    'FixedBatchNormalization',
    'GroupNormalization',
    'LocalResponseNormalization',
    'NormalizeL2',

    # Pooling
    'AveragePooling2D',
    'AveragePoolingND',
    'MaxPooling2D',
    'MaxPoolingND',
    'ROIPooling2D',
    'Unpooling2D',
}

_converters = None


def _get_converters():
    global _converters

    if _converters is not None:
        return _converters

    _converters = {
        name: FunctionConverter(getattr(functions, 'convert_'+name, None))
        for name in _supported_function_node_set}
    return _converters


converters = _get_converters()



import numpy
import six

import chainer
from chainer.backends import _chainerx
from chainer.backends import _cpu
from chainer.backends import cuda
from chainer.backends import intel64
import chainerx

# Aliases
from chainer._backend import Device
from chainer.backends._chainerx import ChainerxDevice
from chainer.backends._chainerx import from_chx  # NOQA
from chainer.backends._chainerx import to_chx  # NOQA
from chainer.backends._cpu import CpuDevice
from chainer.backends.cuda import GpuDevice
from chainer.backends.intel64 import Intel64Device
from chainer import types  # NOQA


def _contains_nan(x):
    """Returns whether the input array has NaN values.

    Args:
        x (numpy.ndarray or cupy.ndarray): Array to be checked.

    Returns:
        bool: True if the input has NaN values.

    """
    if x.dtype.kind in ('f', 'c'):
        device = get_device_from_array(x)
        with chainer.using_device(device):
            return device.xp.isnan(x).any()
    else:
        return False


def copyto(dst, src):
    """Copies the elements of an ndarray to those of another one.

    This function can copy the CPU/GPU arrays to the destination arrays on
    another device.

    Args:
        dst (:class:`numpy.ndarray`, :class:`cupy.ndarray`, \
        :class:`ideep4py.mdarray` or :class:`chainerx.ndarray`):
            Destination array.
        src (:class:`numpy.ndarray`, :class:`cupy.ndarray`, \
        :class:`ideep4py.mdarray` or :class:`chainerx.ndarray`):
            Source array.

    """
    if isinstance(dst, chainerx.ndarray):
        dst[...] = _chainerx._array_to_chainerx(src, dst.device)
        return

    if isinstance(src, chainerx.ndarray):
        src = from_chx(src)

    if isinstance(dst, numpy.ndarray):
        numpy.copyto(dst, _cpu._to_cpu(src))
    elif isinstance(dst, intel64.mdarray):
        intel64.ideep.basic_copyto(
            dst, _cpu._to_cpu(src))
    elif isinstance(dst, cuda.ndarray):
        if isinstance(src, chainer.get_cpu_array_types()):
            src = numpy.asarray(src)
            if dst.flags.c_contiguous or dst.flags.f_contiguous:
                dst.set(src)
            else:
                cuda.cupy.copyto(dst, cuda.to_gpu(src, device=dst.device))
        elif isinstance(src, cuda.ndarray):
            cuda.cupy.copyto(dst, src)
        else:
            raise TypeError('cannot copy from non-array object of type {}'
                            .format(type(src)))
    else:
        raise TypeError('cannot copy to non-array object of type {}'.format(
            type(dst)))


def _guess_device_from_array_module(xp):
    """Returns a plausible device from array module

    .. warning::

        There can be multiple devices for a module

    """
    if xp is cuda.cupy:
        return cuda.GpuDevice(cuda.Device())
    elif xp is chainerx:
        return _chainerx.ChainerxDevice(chainerx.get_default_device())
    else:
        # Cannot detect intel64, because xp of intel64 is numpy.
        return _cpu.CpuDevice()


def get_device(device_spec: types.DeviceSpec) -> Device:
    """Returns a device object.

    Args:
        device_spec (object): Device specifier.
            If a :class:`chainer.backend.Device` instance is given, it is
            returned intact. Otherwise the following values are supported:

            * ChainerX devices

              * A string representing a device.
                (ex. ``'native:0'``, ``'native'``)
              * A :class:`chainerx.Device` object.

            * CuPy

              * A string starts with ``'@cupy:'``.
                (ex. ``'@cupy:0'``)
              * A :class:`cupy.cuda.Device` object.

            * NumPy

              * The string ``'@numpy'``.

            * NumPy with Intel Architecture

              * The string ``'@intel64'``.
    """
    if isinstance(device_spec, Device):
        return device_spec

    if isinstance(device_spec, cuda._integer_types):
        return _get_device_cupy_or_numpy(device_spec)

    if chainerx.is_available() and isinstance(device_spec, chainerx.Device):
        return _chainerx.ChainerxDevice(device_spec)

    if cuda.available and isinstance(device_spec, cuda.Device):
        return cuda.GpuDevice(device_spec)

    if isinstance(device_spec, six.string_types):
        # '-1', '0', '1', ...
        try:
            int_device_spec = int(device_spec)
        except ValueError:
            pass
        else:
            return _get_device_cupy_or_numpy(int_device_spec)

        if device_spec.startswith('@'):
            # '@module:...'
            mod_name, colon, precise_spec = device_spec[1:].partition(':')
            if mod_name == 'numpy':
                if not colon:
                    return _cpu.CpuDevice()
            elif mod_name == 'cupy':
                if colon:
                    return cuda.GpuDevice.from_device_id(int(precise_spec))
            elif mod_name == 'intel64':
                if not colon:
                    return intel64.Intel64Device()
            raise ValueError(
                'Device specifiers starting with \'@\' must be followed by'
                ' a module name and depending on the module, module specific'
                ' precise device specifiers. Actual: {}'.format(device_spec))
        else:
            # String device specifier without '@' prefix is assumed to be a
            # ChainerX device.
            if not chainerx.is_available():
                raise RuntimeError(
                    'Tried to parse ChainerX device specifier \'{}\', '
                    'but ChainerX is not available. '
                    'Note that device specifiers without \'@\' prefix are '
                    'assumed to be ChainerX device '
                    'specifiers.'.format(device_spec))
            return _chainerx.ChainerxDevice(chainerx.get_device(device_spec))

    raise TypeError(
        'Device specifier must be a backend.Device, cuda.Device,'
        ' chainerx.Device, integer or a string. Actual: {}'.format(
            type(device_spec)))


def _get_device_cupy_or_numpy(device_spec):
    # legacy spec of (gpu) device
    if device_spec >= 0:
        return cuda.GpuDevice.from_device_id(device_spec)
    else:
        return _cpu.CpuDevice()


def using_device(device_spec):
    """Context manager to apply the thread-local device state.

    Args:
        device_spec (object): Device specifier. See :func:`chainer.get_device`
            for details.

    .. admonition:: Example

        .. testcode::
           :skipif: doctest_helper.skipif_not_enough_cuda_devices(2)

           with chainer.using_device('@cupy:1'):
               a = cupy.empty((3, 2))

           assert a.device.id == 1

    """

    # TODO(niboshi): Set default device (once this concept is introduced in
    # Chainer).
    device = get_device(device_spec)
    return device.create_context()


def get_array_module(*args):
    """Gets an appropriate NumPy-compatible module to process arguments

    This function will return their data arrays' array module for
    :class:`~chainer.Variable` arguments.

    Args:
        args: Values to determine whether NumPy, CuPy, or ChainerX should be
            used.

    Returns:
        module: :mod:`numpy`, :mod:`cupy`, or :mod:`chainerx` is returned based
        on the types of the arguments.

    """
    is_chainerx_available = chainerx.is_available()
    if is_chainerx_available or cuda.available:
        arrays = []
        for arg in args:
            # Unwrap arrays
            if isinstance(arg, chainer.variable.Variable):
                array = arg.data
            else:
                array = arg
            if is_chainerx_available and isinstance(array, chainerx.ndarray):
                return chainerx
            arrays.append(array)
        if cuda.available:
            return cuda.cupy.get_array_module(*arrays)
    return numpy


def get_device_from_array(*arrays):
    """Gets the device from arrays.

    The device on which the given array reside is returned.

    .. note::

        Unlike :func:`get_array_module`, this method does not recognize
        :class:`~chainer.Variable` objects.
        If you need to get device from the :class:`~chainer.Variable` instance
        ``v``, you need to use ``get_device_from_array(v.array)``.

    Args:
        arrays (array or list of arrays):
            Arrays to determine the device. If multiple arrays are given, the
            device correspoinding to the first array which is not NumPy array
            is returned.

    Returns:
        chainer.backend.Device: Device instance.
    """
    for array in arrays:
        device = GpuDevice.from_array(array)
        if device is not None:
            return device

        if isinstance(array, chainerx.ndarray):
            return ChainerxDevice(array.device)

        device = Intel64Device.from_array(array)
        if device is not None:
            return device

    return CpuDevice()



import onnx


__func_name = None  # not care the name is unique on whole graph


def set_func_name(func_name):
    """Set the name of Chainer function being converted.

    Args:
        func_name (str): The name of Chainer function.
    """
    global __func_name
    __func_name = func_name


def get_func_name():
    """Return processing function name

    """
    assert __func_name is not None
    return __func_name


def make_node(*args, **kwargs):
    """A thin wrapper of `onnx.helper.make_node`.

    Node name will be assigned automatically.

    Args:
        *args (tuple): ONNX node parameters of the node
        **kwargs (dict): ONNX attributes of the node.
    Returns:
        An `onnx.NodeProto` object.
    """
    return onnx.helper.make_node(*args, name=get_func_name(), **kwargs)


class GraphBuilder(object):
    """A helper class to build consecutive ONNX nodes."""

    def __init__(self):
        self._nodes = []
        self._func_name = get_func_name()

    def node_name(self):
        return '{}_tmp_{}'.format(self._func_name, len(self._nodes))

    def op(self, op_name, input_names, num_outputs=1, **kwargs):
        """Creates a new ONNX node and returns its outputs.

        Args:
            op_name (str): The name of an ONNX op.
            input_names (list of str): The names of input values.
            num_outputs (int): The number of output values.
            **kwargs (dict): ONNX attributes of the node.

        Returns:
            A str of the output name when `num_outputs` is 1.
            A tuple of str of the output names otherwise.
        """
        if num_outputs == 1:
            output_names = [self.node_name()]
        else:
            output_names = ['{}_{}'.format(self.node_name(), i) for
                            i in range(num_outputs)]
        return self.op_output_named(
            op_name, input_names, output_names, **kwargs)

    def op_output_named(
            self, op_name, input_names, output_names, **kwargs):
        """Creates a new ONNX node with output names, and returns its outputs.

        Args:
            op_name (str): The name of an ONNX op.
            input_names (list of str): The names of input values.
            output_names (int of str): The names of output values.
            **kwargs (dict): ONNX attributes of the node.

        Returns:
            A str of the output name when number of output is 1.
            A tuple of str of the output names otherwise.
        """
        # Prevent a common mistake. `input_names="input"` creates a
        # node with 5 inputs.
        assert not isinstance(input_names, str)
        node = onnx.helper.make_node(
            op_name, input_names, output_names, name=self.node_name(),
            **kwargs)
        self._nodes.append(node)
        if len(output_names) == 1:
            return node.output[0]
        else:
            return tuple(node.output)

    def nodes(self, output_names=None):
        """Returns all nodes created so far.

        Args:
            output_names (list of str): The names of output values to be set at
                the last node.

        Returns:
            A list of `onnx.NodeProto` objects, suitable as the return
            value of converter functions.
        """
        if output_names is not None:
            assert len(self._nodes[-1].output) == len(output_names)
            self._nodes[-1].output[:] = output_names
        return tuple(self._nodes)


def write_tensor_pb(filename, name, value):
    with open(filename, 'wb') as f:
        t = onnx.numpy_helper.from_array(value, name)
        f.write(t.SerializeToString())


def cleanse_param_name(name):
    """Converts Chainer parameter names to ONNX names.

    Note ONNX identifiers must be a valid C identifier.

    Args:
        name (str): A Chainer parameter name (e.g., /l/W).

    Returns
        A valid ONNX name (e.g., param_l_W).
    """
    return 'param' + name.replace('/', '_')


def is_support_non_standard_domain():
    # from ONNX 1.5, skip schema check on ops in non-standard domain
    # see: https://github.com/onnx/onnx/pull/1876
    # this checker expects onnx adapts semantic versioning
    versions = onnx.__version__.split('.')
    if len(versions) < 2 or (not versions[1].isdecimal()):
        raise RuntimeError(
            'ONNX-Chainer cannot get major and minor version ONNX module: '
            '{}'.format(onnx.__version__))
    major, minor = versions[0], versions[1]
    return major == '1' and int(minor) >= 5



import chainer

import onnx
from onnx import numpy_helper

from onnx_chainer import onnx_helper


def _tensor_from_array_for_constant(array, name):
    tensor = numpy_helper.from_array(array, name=name)
    # Avoid `raw_data` for better debuggability. This would be OK
    # since constants are usually small.
    field_name = onnx.mapping.STORAGE_TENSOR_TYPE_TO_FIELD.get(
        tensor.data_type, None)
    if field_name is not None:
        tensor.ClearField('raw_data')
        getattr(tensor, field_name)[:] = array.flatten().tolist()
    return tensor


class Context(object):
    """Context of converter

    This context shares names during exporting.

    Attributes:
        name_list (dict): list of being exported as ONNX node name with pinned
            or not, keyed by instance ID. When the target variable is
            ``chainer.Variable`` or ``chainer.Parameter``, instance ID of
            ``ndarray`` held by the variable is also put as key, because some
            functions like ``F.where`` internally unwrap variable.

    """

    def __init__(self, model):
        self.name_list = dict()
        self.parameters = []
        self.constants = []
        self.implicit_inputs = dict()  # inputs which not connect to output
        namedlink = {n: l for n, l in model.namedlinks()}
        self.param_to_link = {}
        for name, param in model.namedparams():
            owned_link_name = name[:name.rindex('/')]
            if owned_link_name in namedlink:
                onnx_owned_link_name = onnx_helper.cleanse_param_name(
                    owned_link_name)
                self.param_to_link[id(param)] = (
                    onnx_owned_link_name, namedlink[owned_link_name])
            onnx_name = onnx_helper.cleanse_param_name(name)
            self.set_name(param, onnx_name)

    def get_name(self, variable):
        str_id = id(variable)
        if str_id in self.name_list:
            return self.name_list[str_id][0]
        else:
            new_name = 'v{}'.format(len(self.name_list))
            self.set_name(variable, new_name)
            return new_name

    def set_name(self, variable, name, pinned=False):
        """Set ONNX node name

        Arguments:
            variable (var): target variable
            name (str): name to be exported as ONNX node name
            pinned (bool): if ``True``, the name will not be overwritten in
                subsequence process.
        """

        str_id = id(variable)
        assert str_id not in self.name_list or not self.name_list[str_id][1]
        self.name_list[str_id] = (name, pinned)
        if isinstance(variable, (chainer.Variable, chainer.Parameter)):
            array_id = id(variable.array)
            self.name_list[array_id] = (name, pinned)

    def is_pinned(self, variable):
        str_id = id(variable)
        if str_id not in self.name_list:
            return False
        return self.name_list[str_id][1]

    def add_param(self, array, name, use_original_name=False):
        """Add a parameter array as an ONNX initializer.

        Returns:
            str: registered name.
        """
        if use_original_name:
            onnx_name = name
        else:
            if not (name.startswith('/') or name.startswith('_')):
                name = '/' + name
            onnx_name = '{}_{}'.format(
                onnx_helper.get_func_name(),
                onnx_helper.cleanse_param_name(name))
        self.set_name(array, onnx_name)
        self.parameters.append(array)
        return onnx_name

    def add_const(self, array, name):
        """Add a constant array as an ONNX Constant node.

        Returns:
            str: registered name.
        """
        assert '/' not in name
        onnx_name = '{}_const_{}'.format(onnx_helper.get_func_name(), name)
        self.set_name(array, onnx_name)
        tensor = _tensor_from_array_for_constant(array, name=onnx_name)
        const_node = onnx_helper.make_node(
            'Constant', [], [onnx_name], value=tensor)
        self.constants.append(const_node)
        return onnx_name

    def get_link(self, param):
        """Return link with name which has the param.

        Arguments:
            param(chainer.Parameter): the target param.

        Returns:
            tuple: name and link. returns ``None`` when not found.
        """
        return self.param_to_link.get(id(param), None)
